"""
Manage the Windows registry

-----
Hives
-----
Hives are the main sections of the registry and all begin with the word HKEY.

    - HKEY_LOCAL_MACHINE
    - HKEY_CURRENT_USER
    - HKEY_USER

----
Keys
----
Keys are the folders in the registry. Keys can have many nested subkeys. Keys
can have a value assigned to them under the (Default) value name

-----------------
Values or Entries
-----------------
Values/Entries are name/data pairs. There can be many values in a key. The
(Default) value corresponds to the Key itself, the rest are their own name/value
pairs.

:depends:   - PyWin32
"""
# When production windows installer is using Python 3, Python 2 code can be removed

import logging

import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

try:
    import win32api
    import win32con
    import win32gui

    HAS_WINDOWS_MODULES = True
except ImportError:
    HAS_WINDOWS_MODULES = False


log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "reg"


def __virtual__():
    """
    Only works on Windows systems with the PyWin32
    """
    if not salt.utils.platform.is_windows():
        return (
            False,
            "reg execution module failed to load: "
            "The module will only run on Windows systems",
        )

    if not HAS_WINDOWS_MODULES:
        return (
            False,
            "reg execution module failed to load: "
            "One of the following libraries did not load: "
            "win32gui, win32con, win32api",
        )

    return __virtualname__


def _to_mbcs(vdata):
    """
    Converts unicode to to current users character encoding. Use this for values
    returned by reg functions
    """
    return salt.utils.stringutils.to_unicode(vdata, "mbcs")


def _to_unicode(vdata):
    """
    Converts from current users character encoding to unicode. Use this for
    parameters being pass to reg functions
    """
    # None does not convert to Unicode
    if vdata is None:
        vdata = ""
    if isinstance(vdata, int):
        vdata = str(vdata)
    return salt.utils.stringutils.to_unicode(vdata, "utf-8")


class Registry:  # pylint: disable=R0903
    """
    This was put in a class to delay usage until this module is actually used
    This class contains all the lookup dicts for working with the registry
    """

    def __init__(self):
        self.hkeys = {
            "HKEY_CURRENT_CONFIG": win32con.HKEY_CURRENT_CONFIG,
            "HKEY_CLASSES_ROOT": win32con.HKEY_CLASSES_ROOT,
            "HKEY_CURRENT_USER": win32con.HKEY_CURRENT_USER,
            "HKEY_LOCAL_MACHINE": win32con.HKEY_LOCAL_MACHINE,
            "HKEY_USERS": win32con.HKEY_USERS,
            "HKCC": win32con.HKEY_CURRENT_CONFIG,
            "HKCR": win32con.HKEY_CLASSES_ROOT,
            "HKCU": win32con.HKEY_CURRENT_USER,
            "HKLM": win32con.HKEY_LOCAL_MACHINE,
            "HKU": win32con.HKEY_USERS,
        }
        self.vtype = {
            "REG_NONE": 0,
            "REG_BINARY": win32con.REG_BINARY,
            "REG_DWORD": win32con.REG_DWORD,
            "REG_EXPAND_SZ": win32con.REG_EXPAND_SZ,
            "REG_MULTI_SZ": win32con.REG_MULTI_SZ,
            "REG_SZ": win32con.REG_SZ,
            "REG_QWORD": win32con.REG_QWORD,
        }
        self.opttype = {"REG_OPTION_NON_VOLATILE": 0, "REG_OPTION_VOLATILE": 1}
        # Return Unicode due to from __future__ import unicode_literals
        self.vtype_reverse = {
            0: "REG_NONE",
            win32con.REG_BINARY: "REG_BINARY",
            win32con.REG_DWORD: "REG_DWORD",
            win32con.REG_EXPAND_SZ: "REG_EXPAND_SZ",
            win32con.REG_MULTI_SZ: "REG_MULTI_SZ",
            win32con.REG_SZ: "REG_SZ",
            win32con.REG_QWORD: "REG_QWORD",
        }
        self.opttype_reverse = {0: "REG_OPTION_NON_VOLATILE", 1: "REG_OPTION_VOLATILE"}
        # delete_key_recursive uses this to check the subkey contains enough \
        # as we do not want to remove all or most of the registry
        self.subkey_slash_check = {
            win32con.HKEY_CURRENT_USER: 0,
            win32con.HKEY_LOCAL_MACHINE: 1,
            win32con.HKEY_USERS: 1,
            win32con.HKEY_CURRENT_CONFIG: 1,
            win32con.HKEY_CLASSES_ROOT: 1,
        }

        self.registry_32 = {
            True: win32con.KEY_READ | win32con.KEY_WOW64_32KEY,
            False: win32con.KEY_READ,
        }

    def __getattr__(self, k):
        try:
            return self.hkeys[k]
        except KeyError:
            msg = "No hkey named '{0}. Try one of {1}'"
            hkeys = ", ".join(self.hkeys)
            raise CommandExecutionError(msg.format(k, hkeys))


def key_exists(hive, key, use_32bit_registry=False):
    """
    Check that the key is found in the registry. This refers to keys and not
    value/data pairs. To check value/data pairs, use ``value_exists``

    Args:

        hive (str): The hive to connect to

        key (str): The key to check

        use_32bit_registry (bool): Look in the 32bit portion of the registry

    Returns:
        bool: True if exists, otherwise False

    Usage:

        .. code-block:: python

            import salt.utils.win_reg as reg
            reg.key_exists(hive='HKLM', key='SOFTWARE\\Microsoft')
    """
    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    access_mask = registry.registry_32[use_32bit_registry]

    handle = None
    try:
        handle = win32api.RegOpenKeyEx(hkey, local_key, 0, access_mask)
        return True
    except win32api.error as exc:
        if exc.winerror == 2:
            return False
        if exc.winerror == 5:
            # It exists, but we don't have permission to read it
            return True
        raise
    finally:
        if handle:
            win32api.RegCloseKey(handle)


def value_exists(hive, key, vname, use_32bit_registry=False):
    """
    Check that the value/data pair is found in the registry.

    .. versionadded:: 2018.3.4

    Args:

        hive (str): The hive to connect to

        key (str): The key to check in

        vname (str): The name of the value/data pair you're checking

        use_32bit_registry (bool): Look in the 32bit portion of the registry

    Returns:
        bool: True if exists, otherwise False

    Usage:

        .. code-block:: python

            import salt.utils.win_reg as reg
            reg.value_exists(hive='HKLM',
                                key='SOFTWARE\\Microsoft\\Windows\\CurrentVersion',
                                vname='CommonFilesDir')
    """
    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)
    local_vname = _to_unicode(vname)

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    access_mask = registry.registry_32[use_32bit_registry]

    try:
        handle = win32api.RegOpenKeyEx(hkey, local_key, 0, access_mask)
    except win32api.error as exc:
        if exc.winerror == 2:
            # The key containing the value/data pair does not exist
            return False
        raise

    try:
        # RegQueryValueEx returns and accepts unicode data
        _, _ = win32api.RegQueryValueEx(handle, local_vname)
        # value/data pair exists
        return True
    except win32api.error as exc:
        if exc.winerror == 2 and vname is None:
            # value/data pair exists but is empty
            return True
        else:
            # value/data pair not found
            return False
    finally:
        if handle:
            win32api.RegCloseKey(handle)


def broadcast_change():
    """
    Refresh the windows environment.

    .. note::
        This will only effect new processes and windows. Services will not see
        the change until the system restarts.

    Returns:
        bool: True if successful, otherwise False

    Usage:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.broadcast_change()
    """
    # https://msdn.microsoft.com/en-us/library/windows/desktop/ms644952(v=vs.85).aspx
    _, res = win32gui.SendMessageTimeout(
        win32con.HWND_BROADCAST,
        win32con.WM_SETTINGCHANGE,
        0,
        0,
        win32con.SMTO_ABORTIFHUNG,
        5000,
    )
    return not bool(res)


def list_keys(hive, key=None, use_32bit_registry=False):
    """
    Enumerates the subkeys in a registry key or hive.

    Args:

       hive (str):
            The name of the hive. Can be one of the following:

                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_CURRENT_USER or HKCU
                - HKEY_USERS or HKU
                - HKEY_CLASSES_ROOT or HKCR
                - HKEY_CURRENT_CONFIG or HKCC

        key (str):
            The key (looks like a path) to the value name. If a key is not
            passed, the keys under the hive will be returned.

        use_32bit_registry (bool):
            Accesses the 32bit portion of the registry on 64 bit installations.
            On 32bit machines this is ignored.

    Returns:
        list: A list of keys/subkeys under the hive or key.

    Usage:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.list_keys(hive='HKLM', key='SOFTWARE\\Microsoft')
    """

    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    access_mask = registry.registry_32[use_32bit_registry]

    subkeys = []
    handle = None
    try:
        handle = win32api.RegOpenKeyEx(hkey, local_key, 0, access_mask)

        for i in range(win32api.RegQueryInfoKey(handle)[0]):
            subkey = win32api.RegEnumKey(handle, i)
            subkeys.append(subkey)

    except win32api.error as exc:
        if exc.winerror == 2:
            log.debug(r"Cannot find key: %s\%s", hive, key, exc_info=True)
            return False, r"Cannot find key: {}\{}".format(hive, key)
        if exc.winerror == 5:
            log.debug(r"Access is denied: %s\%s", hive, key, exc_info=True)
            return False, r"Access is denied: {}\{}".format(hive, key)
        raise

    finally:
        if handle:
            win32api.RegCloseKey(handle)

    return subkeys


def list_values(hive, key=None, use_32bit_registry=False):
    """
    Enumerates the values in a registry key or hive.

    .. note::
        The ``(Default)`` value will only be returned if it is set, otherwise it
        will not be returned in the list of values.

    Args:

        hive (str):
            The name of the hive. Can be one of the following:

                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_CURRENT_USER or HKCU
                - HKEY_USER or HKU
                - HKEY_CLASSES_ROOT or HKCR
                - HKEY_CURRENT_CONFIG or HKCC

        key (str):
            The key (looks like a path) to the value name. If a key is not
            passed, the values under the hive will be returned.

        use_32bit_registry (bool):
            Accesses the 32bit portion of the registry on 64 bit installations.
            On 32bit machines this is ignored.

    Returns:
        list: A list of values under the hive or key.

    Usage:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.list_values(hive='HKLM', key='SYSTEM\\CurrentControlSet\\Services\\Tcpip')
    """
    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    access_mask = registry.registry_32[use_32bit_registry]
    handle = None
    values = list()

    try:
        handle = win32api.RegOpenKeyEx(hkey, local_key, 0, access_mask)

        for i in range(win32api.RegQueryInfoKey(handle)[1]):
            vname, vdata, vtype = win32api.RegEnumValue(handle, i)

            if not vname:
                vname = "(Default)"

            value = {
                "hive": local_hive,
                "key": local_key,
                "vname": _to_mbcs(vname),
                "vtype": registry.vtype_reverse[vtype],
                "success": True,
            }
            # Only convert text types to unicode
            if vtype == win32con.REG_MULTI_SZ:
                value["vdata"] = [_to_mbcs(i) for i in vdata]
            elif vtype in [win32con.REG_SZ, win32con.REG_EXPAND_SZ]:
                value["vdata"] = _to_mbcs(vdata)
            else:
                value["vdata"] = vdata
            values.append(value)

    except win32api.error as exc:
        if exc.winerror == 2:
            log.debug(r"Cannot find key: %s\%s", hive, key)
            return False, r"Cannot find key: {}\{}".format(hive, key)
        elif exc.winerror == 5:
            log.debug(r"Access is denied: %s\%s", hive, key)
            return False, r"Access is denied: {}\{}".format(hive, key)
        raise

    finally:
        if handle:
            win32api.RegCloseKey(handle)
    return values


def read_value(hive, key, vname=None, use_32bit_registry=False):
    r"""
    Reads a registry value entry or the default value for a key. To read the
    default value, don't pass ``vname``

    Args:

        hive (str): The name of the hive. Can be one of the following:

            - HKEY_LOCAL_MACHINE or HKLM
            - HKEY_CURRENT_USER or HKCU
            - HKEY_USER or HKU
            - HKEY_CLASSES_ROOT or HKCR
            - HKEY_CURRENT_CONFIG or HKCC

        key (str):
            The key (looks like a path) to the value name.

        vname (str):
            The value name. These are the individual name/data pairs under the
            key. If not passed, the key (Default) value will be returned.

        use_32bit_registry (bool):
            Accesses the 32bit portion of the registry on 64bit installations.
            On 32bit machines this is ignored.

    Returns:
        dict: A dictionary containing the passed settings as well as the
            value_data if successful. If unsuccessful, sets success to False.

        bool: Returns False if the key is not found

        If vname is not passed:

            - Returns the first unnamed value (Default) as a string.
            - Returns none if first unnamed value is empty.

    Usage:

        The following will get the value of the ``version`` value name in the
        ``HKEY_LOCAL_MACHINE\\SOFTWARE\\Salt`` key

        .. code-block:: python

            import salt.utils.win_reg as reg
            reg.read_value(hive='HKLM', key='SOFTWARE\\Salt', vname='version')

    Usage:

        The following will get the default value of the
        ``HKEY_LOCAL_MACHINE\\SOFTWARE\\Salt`` key

        .. code-block:: python

            import salt.utils.win_reg as reg
            reg.read_value(hive='HKLM', key='SOFTWARE\\Salt')
    """
    # If no name is passed, the default value of the key will be returned
    # The value name is Default

    # Setup the return array
    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)
    local_vname = _to_unicode(vname)

    ret = {
        "hive": local_hive,
        "key": local_key,
        "vname": local_vname,
        "vdata": None,
        "vtype": None,
        "success": True,
    }

    if not vname:
        ret["vname"] = "(Default)"

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    access_mask = registry.registry_32[use_32bit_registry]

    handle = None
    try:
        handle = win32api.RegOpenKeyEx(hkey, local_key, 0, access_mask)
        try:
            # RegQueryValueEx returns and accepts unicode data
            vdata, vtype = win32api.RegQueryValueEx(handle, local_vname)
            if vdata or vdata in [0, "", []]:
                # Only convert text types to unicode
                ret["vtype"] = registry.vtype_reverse[vtype]
                if vtype == win32con.REG_MULTI_SZ:
                    ret["vdata"] = [_to_mbcs(i) for i in vdata]
                elif vtype in [win32con.REG_SZ, win32con.REG_EXPAND_SZ]:
                    ret["vdata"] = _to_mbcs(vdata)
                else:
                    ret["vdata"] = vdata
            else:
                ret["comment"] = "Empty Value"
        except win32api.error as exc:
            if exc.winerror == 2 and vname is None:
                ret["vdata"] = "(value not set)"
                ret["vtype"] = "REG_SZ"
            elif exc.winerror == 2:
                msg = "Cannot find {} in {}\\{}".format(
                    local_vname, local_hive, local_key
                )
                log.trace(exc)
                log.trace(msg)
                ret["comment"] = msg
                ret["success"] = False
            else:
                raise
    except win32api.error as exc:
        if exc.winerror == 2:
            msg = "Cannot find key: {}\\{}".format(local_hive, local_key)
            log.trace(exc)
            log.trace(msg)
            ret["comment"] = msg
            ret["success"] = False
        elif exc.winerror == 5:
            msg = "Access is denied: {}\\{}".format(local_hive, local_key)
            log.trace(exc)
            log.trace(msg)
            ret["comment"] = msg
            ret["success"] = False
        else:
            raise
    finally:
        if handle:
            win32api.RegCloseKey(handle)

    return ret


def set_value(
    hive,
    key,
    vname=None,
    vdata=None,
    vtype="REG_SZ",
    use_32bit_registry=False,
    volatile=False,
):
    """
    Sets a value in the registry. If ``vname`` is passed, it will be the value
    for that value name, otherwise it will be the default value for the
    specified key

    Args:

        hive (str):
            The name of the hive. Can be one of the following

                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_CURRENT_USER or HKCU
                - HKEY_USER or HKU
                - HKEY_CLASSES_ROOT or HKCR
                - HKEY_CURRENT_CONFIG or HKCC

        key (str):
            The key (looks like a path) to the value name.

        vname (str):
            The value name. These are the individual name/data pairs under the
            key. If not passed, the key (Default) value will be set.

        vdata (str, int, list, bytes):
            The value you'd like to set. If a value name (vname) is passed, this
            will be the data for that value name. If not, this will be the
            (Default) value for the key.

            The type of data this parameter expects is determined by the value
            type specified in ``vtype``. The correspondence is as follows:

                - REG_BINARY: Binary data (bytes)
                - REG_DWORD: int
                - REG_EXPAND_SZ: str
                - REG_MULTI_SZ: list of str
                - REG_QWORD: int
                - REG_SZ: str

                .. note::
                    When setting REG_BINARY, string data will be converted to
                    binary. You can pass base64 encoded using the ``binascii``
                    built-in module. Use ``binascii.b2a_base64('your data')``

            .. note::
                The type for the (Default) value is always REG_SZ and cannot be
                changed.

            .. note::
                This parameter is optional. If not passed, the Key will be
                created with no associated item/value pairs.

        vtype (str):
            The value type. The possible values of the vtype parameter are
            indicated above in the description of the vdata parameter.

        use_32bit_registry (bool):
            Sets the 32bit portion of the registry on 64bit installations. On
            32bit machines this is ignored.

        volatile (bool):
            When this parameter has a value of True, the registry key will be
            made volatile (i.e. it will not persist beyond a system reset or
            shutdown). This parameter only has an effect when a key is being
            created and at no other time.

    Returns:
        bool: True if successful, otherwise False

    Usage:

        This will set the version value to 2015.5.2 in the SOFTWARE\\Salt key in
        the HKEY_LOCAL_MACHINE hive

        .. code-block:: python

            import salt.utils.win_reg
            winreg.set_value(hive='HKLM', key='SOFTWARE\\Salt', vname='version', vdata='2015.5.2')

    Usage:

        This function is strict about the type of vdata. For instance this
        example will fail because vtype has a value of REG_SZ and vdata has a
        type of int (as opposed to str as expected).

        .. code-block:: python

            import salt.utils.win_reg
            winreg.set_value(hive='HKLM', key='SOFTWARE\\Salt', vname='str_data', vdata=1.2)

    Usage:

        In this next example vdata is properly quoted and should succeed.

        .. code-block:: python

            import salt.utils.win_reg
            winreg.set_value(hive='HKLM', key='SOFTWARE\\Salt', vname='str_data', vdata='1.2')

    Usage:

        This is an example of using vtype REG_BINARY. Both ``set_value``
        commands will set the same value ``Salty Test``

        .. code-block:: python

            import salt.utils.win_reg
            winreg.set_value(hive='HKLM', key='SOFTWARE\\Salt', vname='bin_data', vdata='Salty Test', vtype='REG_BINARY')

            import binascii
            bin_data = binascii.b2a_base64('Salty Test')
            winreg.set_value(hive='HKLM', key='SOFTWARE\\Salt', vname='bin_data_encoded', vdata=bin_data, vtype='REG_BINARY')

    Usage:

        An example using vtype REG_MULTI_SZ is as follows:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.set_value(hive='HKLM', key='SOFTWARE\\Salt', vname='list_data', vdata=['Salt', 'is', 'great'], vtype='REG_MULTI_SZ')
    """
    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)
    local_vname = _to_unicode(vname)
    local_vtype = _to_unicode(vtype)

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    vtype_value = registry.vtype[local_vtype]
    access_mask = registry.registry_32[use_32bit_registry] | win32con.KEY_ALL_ACCESS

    local_vdata = cast_vdata(vdata=vdata, vtype=local_vtype)

    if volatile:
        create_options = registry.opttype["REG_OPTION_VOLATILE"]
    else:
        create_options = registry.opttype["REG_OPTION_NON_VOLATILE"]

    handle = None
    try:
        handle, result = win32api.RegCreateKeyEx(
            hkey, local_key, access_mask, Options=create_options
        )
        msg = (
            "Created new key: %s\\%s" if result == 1 else "Opened existing key: %s\\%s"
        )
        log.debug(msg, local_hive, local_key)

        try:
            win32api.RegSetValueEx(handle, local_vname, 0, vtype_value, local_vdata)
            win32api.RegFlushKey(handle)
            broadcast_change()
            return True
        except TypeError as exc:
            log.exception('"vdata" does not match the expected data type.\n%s', exc)
            return False
        except (SystemError, ValueError) as exc:
            log.exception("Encountered error setting registry value.\n%s", exc)
            return False

    except win32api.error as exc:
        log.exception(
            "Error creating/opening key: %s\\%s\n%s\n%s",
            local_hive,
            local_key,
            exc.winerror,
            exc.strerror,
        )
        return False

    finally:
        if handle:
            win32api.RegCloseKey(handle)


def cast_vdata(vdata=None, vtype="REG_SZ"):
    """
    Cast the ``vdata` value to the appropriate data type for the registry type
    specified in ``vtype``

    Args:

        vdata (str, int, list, bytes): The data to cast

        vtype (str):
            The type of data to be written to the registry. Must be one of the
            following:

                - REG_BINARY
                - REG_DWORD
                - REG_EXPAND_SZ
                - REG_MULTI_SZ
                - REG_QWORD
                - REG_SZ

    Returns:
        The vdata cast to the appropriate type. Will be unicode string, binary,
        list of unicode strings, or int

    Usage:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.cast_vdata(vdata='This is the string', vtype='REG_SZ')
    """
    # Check data type and cast to expected type
    # int will automatically become long on 64bit numbers
    # https://www.python.org/dev/peps/pep-0237/

    registry = Registry()
    vtype_value = registry.vtype[vtype]

    # String Types to Unicode
    if vtype_value in [win32con.REG_SZ, win32con.REG_EXPAND_SZ]:
        return _to_unicode(vdata)
    # Don't touch binary... if it's binary
    elif vtype_value == win32con.REG_BINARY:
        if isinstance(vdata, str):
            # Unicode data must be encoded
            return vdata.encode("utf-8")
        return vdata
    # Make sure REG_MULTI_SZ is a list of strings
    elif vtype_value == win32con.REG_MULTI_SZ:
        return [_to_unicode(i) for i in vdata]
    # Make sure REG_QWORD is a 64 bit integer
    elif vtype_value == win32con.REG_QWORD:
        return int(vdata)
    # Everything else is int
    else:
        return int(vdata)


def delete_key_recursive(hive, key, use_32bit_registry=False):
    """
    .. versionadded:: 2015.5.4

    Delete a registry key to include all subkeys and value/data pairs.

    Args:

        hive (str):
            The name of the hive. Can be one of the following

                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_CURRENT_USER or HKCU
                - HKEY_USER or HKU
                - HKEY_CLASSES_ROOT or HKCR
                - HKEY_CURRENT_CONFIG or HKCC

            key (str):
                The key to remove (looks like a path)

            use_32bit_registry (bool):
                Deletes the 32bit portion of the registry on 64bit
                installations. On 32bit machines this is ignored.

    Returns:
        dict: A dictionary listing the keys that deleted successfully as well as
            those that failed to delete.

    Usage:

        The following example will remove ``salt`` and all its subkeys from the
        ``SOFTWARE`` key in ``HKEY_LOCAL_MACHINE``:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.delete_key_recursive(hive='HKLM', key='SOFTWARE\\DeleteMe')
    """

    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)

    # Instantiate the registry object
    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    key_path = local_key
    access_mask = registry.registry_32[use_32bit_registry] | win32con.KEY_ALL_ACCESS

    if not key_exists(local_hive, local_key, use_32bit_registry):
        log.debug('"%s\\%s" not found', hive, key)
        return False

    if (len(key) > 1) and (key.count("\\", 1) < registry.subkey_slash_check[hkey]):
        log.error('"%s\\%s" is too close to root, not safe to remove', hive, key)
        return False

    # Functions for traversing the registry tree
    def _subkeys(_key):
        """
        Enumerate keys
        """
        i = 0
        while True:
            try:
                subkey = win32api.RegEnumKey(_key, i)
                yield _to_mbcs(subkey)
                i += 1
            except win32api.error:
                break

    def _traverse_registry_tree(_hkey, _keypath, _ret, _access_mask):
        """
        Traverse the registry tree i.e. dive into the tree
        """
        _key = win32api.RegOpenKeyEx(_hkey, _keypath, 0, _access_mask)
        for subkeyname in _subkeys(_key):
            subkeypath = "{}\\{}".format(_keypath, subkeyname)
            _ret = _traverse_registry_tree(_hkey, subkeypath, _ret, access_mask)
            _ret.append(subkeypath)
        return _ret

    # Get a reverse list of registry keys to be deleted
    key_list = []
    key_list = _traverse_registry_tree(hkey, key_path, key_list, access_mask)
    # Add the top level key last, all subkeys must be deleted first
    key_list.append(key_path)

    ret = {"Deleted": [], "Failed": []}

    # Delete all sub_keys
    for sub_key_path in key_list:
        key_handle = None
        try:
            key_handle = win32api.RegOpenKeyEx(hkey, sub_key_path, 0, access_mask)
            try:
                win32api.RegDeleteKey(key_handle, "")
                ret["Deleted"].append(r"{}\{}".format(hive, sub_key_path))
            except OSError as exc:
                log.error(exc, exc_info=True)
                ret["Failed"].append(r"{}\{} {}".format(hive, sub_key_path, exc))
        except win32api.error as exc:
            log.error(exc, exc_info=True)
            ret["Failed"].append(r"{}\{} {}".format(hive, sub_key_path, exc.strerror))
        finally:
            if key_handle:
                win32api.RegCloseKey(key_handle)

    broadcast_change()

    return ret


def delete_value(hive, key, vname=None, use_32bit_registry=False):
    """
    Delete a registry value entry or the default value for a key.

    Args:

        hive (str):
            The name of the hive. Can be one of the following

                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_CURRENT_USER or HKCU
                - HKEY_USER or HKU
                - HKEY_CLASSES_ROOT or HKCR
                - HKEY_CURRENT_CONFIG or HKCC

        key (str):
            The key (looks like a path) to the value name.

        vname (str):
            The value name. These are the individual name/data pairs under the
            key. If not passed, the key (Default) value will be deleted.

        use_32bit_registry (bool):
            Deletes the 32bit portion of the registry on 64bit installations. On
            32bit machines this is ignored.

    Return:
        bool: True if successful, otherwise False

    Usage:

        .. code-block:: python

            import salt.utils.win_reg
            winreg.delete_value(hive='HKLM', key='SOFTWARE\\SaltTest', vname='version')
    """
    local_hive = _to_unicode(hive)
    local_key = _to_unicode(key)
    local_vname = _to_unicode(vname)

    registry = Registry()
    try:
        hkey = registry.hkeys[local_hive]
    except KeyError:
        raise CommandExecutionError("Invalid Hive: {}".format(local_hive))
    access_mask = registry.registry_32[use_32bit_registry] | win32con.KEY_ALL_ACCESS

    handle = None
    try:
        handle = win32api.RegOpenKeyEx(hkey, local_key, 0, access_mask)
        win32api.RegDeleteValue(handle, local_vname)
        broadcast_change()
        return True
    except win32api.error as exc:
        if exc.winerror == 2:
            return None
        raise
    finally:
        if handle:
            win32api.RegCloseKey(handle)



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import salt.modules.nova as nova
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NovaTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.nova
    """

    def setup_loader_modules(self):
        patcher = patch("salt.modules.nova._auth")
        self.mock_auth = patcher.start()
        self.addCleanup(patcher.stop)
        self.addCleanup(delattr, self, "mock_auth")
        return {nova: {}}

    def test_boot(self):
        """
        Test for Boot (create) a new instance
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "boot", MagicMock(return_value="A")):
            self.assertTrue(nova.boot("name"))

    def test_volume_list(self):
        """
        Test for List storage volumes
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "volume_list", MagicMock(return_value="A")):
            self.assertTrue(nova.volume_list())

    def test_volume_show(self):
        """
        Test for Create a block storage volume
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "volume_show", MagicMock(return_value="A")):
            self.assertTrue(nova.volume_show("name"))

    def test_volume_create(self):
        """
        Test for Create a block storage volume
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "volume_create", MagicMock(return_value="A")):
            self.assertTrue(nova.volume_create("name"))

    def test_volume_delete(self):
        """
        Test for Destroy the volume
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "volume_delete", MagicMock(return_value="A")):
            self.assertTrue(nova.volume_delete("name"))

    def test_volume_detach(self):
        """
        Test for Attach a block storage volume
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "volume_detach", MagicMock(return_value="A")):
            self.assertTrue(nova.volume_detach("name"))

    def test_volume_attach(self):
        """
        Test for Attach a block storage volume
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "volume_attach", MagicMock(return_value="A")):
            self.assertTrue(nova.volume_attach("name", "serv_name"))

    def test_suspend(self):
        """
        Test for Suspend an instance
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "suspend", MagicMock(return_value="A")):
            self.assertTrue(nova.suspend("instance_id"))

    def test_resume(self):
        """
        Test for Resume an instance
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "resume", MagicMock(return_value="A")):
            self.assertTrue(nova.resume("instance_id"))

    def test_lock(self):
        """
        Test for Lock an instance
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "lock", MagicMock(return_value="A")):
            self.assertTrue(nova.lock("instance_id"))

    def test_delete(self):
        """
        Test for Delete an instance
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "delete", MagicMock(return_value="A")):
            self.assertTrue(nova.delete("instance_id"))

    def test_flavor_list(self):
        """
        Test for Return a list of available flavors (nova flavor-list)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "flavor_list", MagicMock(return_value="A")):
            self.assertTrue(nova.flavor_list())

    def test_flavor_create(self):
        """
        Test for Add a flavor to nova (nova flavor-create)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "flavor_create", MagicMock(return_value="A")):
            self.assertTrue(nova.flavor_create("name"))

    def test_flavor_delete(self):
        """
        Test for Delete a flavor from nova by id (nova flavor-delete)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "flavor_delete", MagicMock(return_value="A")):
            self.assertTrue(nova.flavor_delete("flavor_id"))

    def test_keypair_list(self):
        """
        Test for Return a list of available keypairs (nova keypair-list)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "keypair_list", MagicMock(return_value="A")):
            self.assertTrue(nova.keypair_list())

    def test_keypair_add(self):
        """
        Test for Add a keypair to nova (nova keypair-add)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "keypair_add", MagicMock(return_value="A")):
            self.assertTrue(nova.keypair_add("name"))

    def test_keypair_delete(self):
        """
        Test for Add a keypair to nova (nova keypair-delete)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "keypair_delete", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.keypair_delete("name"))

    def test_image_list(self):
        """
        Test for Return a list of available images
         (nova images-list + nova image-show)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "image_list", MagicMock(return_value="A")):
            self.assertTrue(nova.image_list())

    def test_image_meta_set(self):
        """
        Test for Sets a key=value pair in the
         metadata for an image (nova image-meta set)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "image_meta_set", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.image_meta_set())

    def test_image_meta_delete(self):
        """
        Test for Delete a key=value pair from the metadata for an image
        (nova image-meta set)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "image_meta_delete", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.image_meta_delete())

    def test_list_(self):
        """
        Test for To maintain the feel of the nova command line,
         this function simply calls
         the server_list function.
        """
        with patch.object(nova, "server_list", return_value=["A"]):
            self.assertEqual(nova.list_(), ["A"])

    def test_server_list(self):
        """
        Test for Return list of active servers
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "server_list", MagicMock(return_value="A")):
            self.assertTrue(nova.server_list())

    def test_show(self):
        """
        Test for To maintain the feel of the nova command line,
         this function simply calls
         the server_show function.
        """
        with patch.object(nova, "server_show", return_value=["A"]):
            self.assertEqual(nova.show("server_id"), ["A"])

    def test_server_list_detailed(self):
        """
        Test for Return detailed list of active servers
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "server_list_detailed", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.server_list_detailed())

    def test_server_show(self):
        """
        Test for Return detailed information for an active server
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "server_show", MagicMock(return_value="A")):
            self.assertTrue(nova.server_show("serv_id"))

    def test_secgroup_create(self):
        """
        Test for Add a secgroup to nova (nova secgroup-create)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "server_list_detailed", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.secgroup_create("name", "desc"))

    def test_secgroup_delete(self):
        """
        Test for Delete a secgroup to nova (nova secgroup-delete)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "secgroup_delete", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.secgroup_delete("name"))

    def test_secgroup_list(self):
        """
        Test for Return a list of available security groups (nova items-list)
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(self.mock_auth, "secgroup_list", MagicMock(return_value="A")):
            self.assertTrue(nova.secgroup_list())

    def test_server_by_name(self):
        """
        Test for Return information about a server
        """
        self.mock_auth.side_effect = MagicMock()
        with patch.object(
            self.mock_auth, "server_by_name", MagicMock(return_value="A")
        ):
            self.assertTrue(nova.server_by_name("name"))



import logging
import os

import salt.version

log = logging.getLogger(__name__)


def test_salt_run_exception_exit(salt_run_cli):
    """
    test that the exitcode is 1 when an exception is
    thrown in a salt runner
    """
    ret = salt_run_cli.run(
        "error.error", "name='Exception'", "message='This is an error.'"
    )
    assert ret.returncode == 1


def test_salt_run_non_exception_exit(salt_run_cli):
    """
    Test standard exitcode and output when runner works.
    """
    ret = salt_run_cli.run("test.stdout_print")
    assert ret.returncode == 0
    assert ret.stdout == 'foo\n"bar"\n'


def test_versions_report(salt_run_cli):
    """
    test salt-run --versions-report
    """
    expected = salt.version.versions_information()
    # sanitize expected of unnnecessary whitespace
    for _, section in expected.items():
        for key in section:
            if isinstance(section[key], str):
                section[key] = section[key].strip()

    ret = salt_run_cli.run("--versions-report")
    assert ret.returncode == 0
    assert ret.stdout
    ret_lines = ret.stdout.split("\n")

    assert ret_lines
    # sanitize lines
    ret_lines = [line.strip() for line in ret_lines]

    for header in expected:
        assert "{}:".format(header) in ret_lines

    ret_dict = {}
    expected_keys = set()
    for line in ret_lines:
        if not line:
            continue
        if line.endswith(":"):
            assert not expected_keys
            current_header = line.rstrip(":")
            assert current_header in expected
            ret_dict[current_header] = {}
            expected_keys = set(expected[current_header].keys())
        else:
            key, *value_list = line.split(":", 1)
            assert value_list
            assert len(value_list) == 1
            value = value_list[0].strip()
            if value == "Not Installed":
                value = None
            ret_dict[current_header][key] = value
            assert key in expected_keys
            expected_keys.remove(key)
    assert not expected_keys
    if os.environ.get("ONEDIR_TESTRUN", "0") == "0":
        # Stop any more testing
        return

    assert "relenv" in ret_dict["Dependency Versions"]
    assert "Salt Extensions" in ret_dict


def test_salt_run_version(salt_run_cli):
    expected = salt.version.__saltstack_version__.formatted_version
    ret = salt_run_cli.run("--version")
    assert f"cli_salt_run.py {expected}\n" == ret.stdout



"""
A module to pull data from Foreman via its API into the Pillar dictionary


Configuring the Foreman ext_pillar
==================================

Set the following Salt config to setup Foreman as external pillar source:

.. code-block:: yaml

  ext_pillar:
    - foreman:
        key: foreman # Nest results within this key
        only: ['hostgroup_name', 'parameters'] # Add only these keys to pillar

  foreman.url: https://example.com/foreman_api
  foreman.user: username # default is admin
  foreman.password: password # default is changeme

The following options are optional:

.. code-block:: yaml

  foreman.api: apiversion # default is 2 (1 is not supported yet)
  foreman.verifyssl: False # default is True
  foreman.certfile: /etc/ssl/certs/mycert.pem # default is None
  foreman.keyfile: /etc/ssl/private/mykey.pem # default is None
  foreman.cafile: /etc/ssl/certs/mycert.ca.pem # default is None
  foreman.lookup_parameters: True # default is True

An alternative would be to use the Foreman modules integrating Salt features
in the Smart Proxy and the webinterface.

Further information can be found on `GitHub <https://github.com/theforeman/foreman_salt>`_.

Module Documentation
====================
"""

import logging

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

__opts__ = {
    "foreman.url": "http://foreman/api",
    "foreman.user": "admin",
    "foreman.password": "changeme",
    "foreman.api": 2,
    "foreman.verifyssl": True,
    "foreman.certfile": None,
    "foreman.keyfile": None,
    "foreman.cafile": None,
    "foreman.lookup_parameters": True,
}


# Set up logging
log = logging.getLogger(__name__)

# Declare virtualname
__virtualname__ = "foreman"


def __virtual__():
    """
    Only return if all the modules are available
    """
    if not HAS_REQUESTS:
        return False
    return __virtualname__


def ext_pillar(minion_id, pillar, key=None, only=()):  # pylint: disable=W0613
    """
    Read pillar data from Foreman via its API.
    """
    url = __opts__["foreman.url"]
    user = __opts__["foreman.user"]
    password = __opts__["foreman.password"]
    api = __opts__["foreman.api"]
    verify = __opts__["foreman.verifyssl"]
    certfile = __opts__["foreman.certfile"]
    keyfile = __opts__["foreman.keyfile"]
    cafile = __opts__["foreman.cafile"]
    lookup_parameters = __opts__["foreman.lookup_parameters"]

    log.info("Querying Foreman at %r for information for %r", url, minion_id)
    try:
        # Foreman API version 1 is currently not supported
        if api != 2:
            log.error(
                "Foreman API v2 is supported only, please specify"
                "version 2 in your Salt master config"
            )
            raise Exception

        headers = {"accept": "version=" + str(api) + ",application/json"}

        if verify and cafile is not None:
            verify = cafile

        resp = requests.get(
            url + "/hosts/" + minion_id,
            auth=(user, password),
            headers=headers,
            verify=verify,
            cert=(certfile, keyfile),
        )
        result = resp.json()

        log.debug("Raw response of the Foreman request is %r", result)

        if lookup_parameters:
            parameters = dict()
            for param in result["all_parameters"]:
                parameters.update({param["name"]: param["value"]})

            result["parameters"] = parameters

        if only:
            result = {k: result[k] for k in only if k in result}

    except Exception:  # pylint: disable=broad-except
        log.exception("Could not fetch host data via Foreman API:")
        return {}

    if key:
        result = {key: result}

    return result



import pytest
from pytestskipmarkers.utils import platform
from saltfactories.utils.functional import MultiStateResult


@pytest.mark.skip_on_windows(reason="Linux test only")
def test_services(install_salt, salt_cli, salt_minion):
    """
    Check if Services are enabled/disabled
    """
    if install_salt.compressed:
        pytest.skip("Skip test on single binary and onedir package")

    if install_salt.distro_id in ("ubuntu", "debian"):
        services_enabled = ["salt-master", "salt-minion", "salt-syndic", "salt-api"]
        services_disabled = []
    elif install_salt.distro_id in ("centos", "redhat", "amzn", "fedora", "photon"):
        services_enabled = []
        services_disabled = ["salt-master", "salt-minion", "salt-syndic", "salt-api"]
    elif platform.is_darwin():
        services_enabled = ["salt-minion"]
        services_disabled = []
    else:
        pytest.fail(f"Don't know how to handle os_family={install_salt.distro_id}")

    for service in services_enabled:
        assert (
            "true"
            in salt_cli.run(
                "service.enabled", service, minion_tgt=salt_minion.id
            ).stdout
        )

    for service in services_disabled:
        assert (
            "true"
            in salt_cli.run(
                "service.disabled", service, minion_tgt=salt_minion.id
            ).stdout
        )



"""
SCP Module
==========

.. versionadded:: 2019.2.0

Module to copy files via `SCP <https://man.openbsd.org/scp>`_
"""

import logging

# Import salt modules

try:
    import paramiko
    import scp

    HAS_SCP = True
except ImportError:
    HAS_SCP = False

__proxyenabled__ = ["*"]
__virtualname__ = "scp"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_SCP:
        return False, "Please install SCP for this modules: pip install scp"
    return __virtualname__


def _select_kwargs(**kwargs):
    paramiko_kwargs = {}
    scp_kwargs = {}
    paramiko_args = __utils__["args.get_function_argspec"](paramiko.SSHClient.connect)[
        0
    ]
    paramiko_args.append("auto_add_policy")
    scp_args = __utils__["args.get_function_argspec"](scp.SCPClient.__init__)[0]
    scp_args.pop(0)  # strip transport arg (it is passed in _prepare_connection)
    for key, val in kwargs.items():
        if key in paramiko_args and val is not None:
            paramiko_kwargs[key] = val
        if key in scp_args and val is not None:
            scp_kwargs[key] = val
    return paramiko_kwargs, scp_kwargs


def _prepare_connection(**kwargs):
    """
    Prepare the underlying SSH connection with the remote target.
    """
    paramiko_kwargs, scp_kwargs = _select_kwargs(**kwargs)
    ssh = paramiko.SSHClient()
    if paramiko_kwargs.pop("auto_add_policy", False):
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(**paramiko_kwargs)
    scp_client = scp.SCPClient(ssh.get_transport(), **scp_kwargs)
    return scp_client


def get(remote_path, local_path="", recursive=False, preserve_times=False, **kwargs):
    """
    Transfer files and directories from remote host to the localhost of the
    Minion.

    remote_path
        Path to retrieve from remote host. Since this is evaluated by scp on the
        remote host, shell wildcards and environment variables may be used.

    recursive: ``False``
        Transfer files and directories recursively.

    preserve_times: ``False``
        Preserve ``mtime`` and ``atime`` of transferred files and directories.

    hostname
        The hostname of the remote device.

    port: ``22``
        The port of the remote device.

    username
        The username required for SSH authentication on the device.

    password
        Used for password authentication. It is also used for private key
        decryption if ``passphrase`` is not given.

    passphrase
        Used for decrypting private keys.

    pkey
        An optional private key to use for authentication.

    key_filename
        The filename, or list of filenames, of optional private key(s) and/or
        certificates to try for authentication.

    timeout
        An optional timeout (in seconds) for the TCP connect.

    socket_timeout: ``10``
        The channel socket timeout in seconds.

    buff_size: ``16384``
        The size of the SCP send buffer.

    allow_agent: ``True``
        Set to ``False`` to disable connecting to the SSH agent.

    look_for_keys: ``True``
        Set to ``False`` to disable searching for discoverable private key
        files in ``~/.ssh/``

    banner_timeout
        An optional timeout (in seconds) to wait for the SSH banner to be
        presented.

    auth_timeout
        An optional timeout (in seconds) to wait for an authentication
        response.

    auto_add_policy: ``False``
        Automatically add the host to the ``known_hosts``.

    CLI Example:

    .. code-block:: bash

        salt '*' scp.get /var/tmp/file /tmp/file hostname=10.10.10.1 auto_add_policy=True
    """
    scp_client = _prepare_connection(**kwargs)
    get_kwargs = {"recursive": recursive, "preserve_times": preserve_times}
    if local_path:
        get_kwargs["local_path"] = local_path
    return scp_client.get(remote_path, **get_kwargs)


def put(
    files,
    remote_path=None,
    recursive=False,
    preserve_times=False,
    saltenv="base",
    **kwargs
):
    """
    Transfer files and directories to remote host.

    files
        A single path or a list of paths to be transferred.

    remote_path
        The path on the remote device where to store the files.

    recursive: ``True``
        Transfer files and directories recursively.

    preserve_times: ``False``
        Preserve ``mtime`` and ``atime`` of transferred files and directories.

    hostname
        The hostname of the remote device.

    port: ``22``
        The port of the remote device.

    username
        The username required for SSH authentication on the device.

    password
        Used for password authentication. It is also used for private key
        decryption if ``passphrase`` is not given.

    passphrase
        Used for decrypting private keys.

    pkey
        An optional private key to use for authentication.

    key_filename
        The filename, or list of filenames, of optional private key(s) and/or
        certificates to try for authentication.

    timeout
        An optional timeout (in seconds) for the TCP connect.

    socket_timeout: ``10``
        The channel socket timeout in seconds.

    buff_size: ``16384``
        The size of the SCP send buffer.

    allow_agent: ``True``
        Set to ``False`` to disable connecting to the SSH agent.

    look_for_keys: ``True``
        Set to ``False`` to disable searching for discoverable private key
        files in ``~/.ssh/``

    banner_timeout
        An optional timeout (in seconds) to wait for the SSH banner to be
        presented.

    auth_timeout
        An optional timeout (in seconds) to wait for an authentication
        response.

    auto_add_policy: ``False``
        Automatically add the host to the ``known_hosts``.

    CLI Example:

    .. code-block:: bash

        salt '*' scp.put /path/to/file /var/tmp/file hostname=server1 auto_add_policy=True
    """
    scp_client = _prepare_connection(**kwargs)
    put_kwargs = {"recursive": recursive, "preserve_times": preserve_times}
    if remote_path:
        put_kwargs["remote_path"] = remote_path
    cached_files = []
    if not isinstance(files, (list, tuple)):
        files = [files]
    for file_ in files:
        cached_file = __salt__["cp.cache_file"](file_, saltenv=saltenv)
        cached_files.append(cached_file)
    return scp_client.put(cached_files, **put_kwargs)



import pytest

import salt.states.group as group
from tests.support.mock import MagicMock, patch

__context__ = {}


def ping():
    ...


@pytest.fixture
def configure_loader_modules():
    return {group: {"__salt__": {"test.ping": ping}, "__opts__": {"test": False}}}


def test_present_with_non_unique_gid():
    with patch(
        "salt.states.group._changes", MagicMock(side_effect=[False, {}, False])
    ), patch.dict(
        group.__salt__,
        {"group.getent": MagicMock(side_effect=[[], [{"name": "salt", "gid": 1}]])},
    ), patch.dict(
        group.__salt__, {"group.add": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.info": MagicMock(return_value={"things": "stuff"})}
    ):
        ret = group.present("salt", gid=1, non_unique=True)
        assert ret == {
            "changes": {"things": "stuff"},
            "comment": "New group salt created",
            "name": "salt",
            "result": True,
        }
        ret = group.present("salt", gid=1, non_unique=False)
        assert ret == {
            "changes": {},
            "comment": "Group salt is not present but gid 1 is already taken by group salt",
            "name": "salt",
            "result": False,
        }


def test_present_with_existing_group_and_non_unique_gid():
    with patch(
        "salt.states.group._changes",
        MagicMock(side_effect=[{"gid": 1}, {}, {"gid": 1}, {"gid": 1}]),
    ), patch.dict(
        group.__salt__,
        {
            "group.getent": MagicMock(
                side_effect=[[{"name": "salt", "gid": 1}], [{"name": "salt", "gid": 1}]]
            )
        },
    ), patch.dict(
        group.__salt__, {"group.add": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.chgid": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.info": MagicMock(return_value={"things": "stuff"})}
    ):
        ret = group.present("salt", gid=1, non_unique=True)
        assert ret == {
            "changes": {"Final": "All changes applied successfully"},
            "comment": "The following group attributes are set to be changed:\ngid: 1\n",
            "name": "salt",
            "result": True,
        }
        ret = group.present("salt", gid=1, non_unique=False)
        assert ret == {
            "changes": {"Failed": {"gid": 1}},
            "comment": "The following group attributes are set to be changed:\n"
            "gid: 1\n"
            "Some changes could not be applied",
            "name": "salt",
            "result": False,
        }



"""
unit tests for the libvirt_events engine
"""
import pytest

import salt.engines.libvirt_events as libvirt_events
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {libvirt_events: {}}


@pytest.fixture
def mock_libvirt():
    with patch("salt.engines.libvirt_events.libvirt") as mock_libvirt:
        mock_libvirt.getVersion.return_value = 2000000
        mock_libvirt.virEventRunDefaultImpl.return_value = -1  # Don't loop for ever
        mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE = 0
        mock_libvirt.VIR_DOMAIN_EVENT_ID_REBOOT = 1
        mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE = 0
        mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_REFRESH = 1
        mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_LIFECYCLE = 0
        mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_UPDATE = 1
        yield mock_libvirt


def test_get_libvirt_enum_string_subprefix(mock_libvirt):
    """
    Make sure the libvirt enum value to string works reliably with
    elements with a sub prefix, eg VIR_PREFIX_SUB_* in this case.
    """
    # Test case with a sub prefix
    mock_libvirt.VIR_PREFIX_NONE = 0
    mock_libvirt.VIR_PREFIX_ONE = 1
    mock_libvirt.VIR_PREFIX_TWO = 2
    mock_libvirt.VIR_PREFIX_SUB_FOO = 0
    mock_libvirt.VIR_PREFIX_SUB_BAR = 1
    mock_libvirt.VIR_PREFIX_SUB_FOOBAR = 2

    assert libvirt_events._get_libvirt_enum_string("VIR_PREFIX_", 2) == "two"


def test_get_libvirt_enum_string_underscores(mock_libvirt):
    """
    Make sure the libvirt enum value to string works reliably and items
    with an underscore aren't confused with sub prefixes.
    """
    mock_libvirt.VIR_PREFIX_FOO = 0
    mock_libvirt.VIR_PREFIX_BAR_FOO = 1

    assert libvirt_events._get_libvirt_enum_string("VIR_PREFIX_", 1) == "bar foo"


def test_get_domain_event_detail(mock_libvirt):
    """
    Test get_domain_event_detail function
    """
    mock_libvirt.VIR_DOMAIN_EVENT_CRASHED_PANICKED = 0
    mock_libvirt.VIR_DOMAIN_EVENT_DEFINED = 0
    mock_libvirt.VIR_DOMAIN_EVENT_UNDEFINED = 1
    mock_libvirt.VIR_DOMAIN_EVENT_CRASHED = 2
    mock_libvirt.VIR_DOMAIN_EVENT_DEFINED_ADDED = 0
    mock_libvirt.VIR_DOMAIN_EVENT_DEFINED_UPDATED = 1

    assert libvirt_events._get_domain_event_detail(1, 2) == ("undefined", "unknown")
    assert libvirt_events._get_domain_event_detail(0, 1) == ("defined", "updated")
    assert libvirt_events._get_domain_event_detail(4, 2) == ("unknown", "unknown")


def test_event_register(mock_libvirt):
    """
    Test that the libvirt_events engine actually registers events catch them and cleans
    before leaving the place.
    """
    mock_libvirt.VIR_NETWORK_EVENT_ID_LIFECYCLE = 1000

    mock_cnx = MagicMock()
    mock_libvirt.openReadOnly.return_value = mock_cnx

    # Don't loop for ever
    mock_libvirt.virEventRunDefaultImpl.return_value = -1

    mock_cnx.networkEventRegisterAny.return_value = 10000

    libvirt_events.start("test:///", "test/prefix")

    # Check that the connection has been opened
    mock_libvirt.openReadOnly.assert_called_once_with("test:///")

    # Check that the connection has been closed
    mock_cnx.close.assert_called_once()

    # Check events registration and deregistration
    mock_cnx.domainEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
        libvirt_events._domain_event_lifecycle_cb,
        {"prefix": "test/prefix", "object": "domain", "event": "lifecycle"},
    )
    mock_cnx.networkEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_NETWORK_EVENT_ID_LIFECYCLE,
        libvirt_events._network_event_lifecycle_cb,
        {"prefix": "test/prefix", "object": "network", "event": "lifecycle"},
    )
    mock_cnx.storagePoolEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE,
        libvirt_events._pool_event_lifecycle_cb,
        {"prefix": "test/prefix", "object": "pool", "event": "lifecycle"},
    )
    mock_cnx.storagePoolEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_REFRESH,
        libvirt_events._pool_event_refresh_cb,
        {"prefix": "test/prefix", "object": "pool", "event": "refresh"},
    )
    mock_cnx.nodeDeviceEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_LIFECYCLE,
        libvirt_events._nodedev_event_lifecycle_cb,
        {"prefix": "test/prefix", "object": "nodedev", "event": "lifecycle"},
    )
    mock_cnx.nodeDeviceEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_UPDATE,
        libvirt_events._nodedev_event_update_cb,
        {"prefix": "test/prefix", "object": "nodedev", "event": "update"},
    )

    # Check that the deregister events are called with the result of register
    mock_cnx.networkEventDeregisterAny.assert_called_with(
        mock_cnx.networkEventRegisterAny.return_value
    )

    # Check that the default 'all' filter actually worked
    counts = {
        obj: len(callback_def)
        for obj, callback_def in libvirt_events.CALLBACK_DEFS.items()
    }
    for obj, count in counts.items():
        register = libvirt_events.REGISTER_FUNCTIONS[obj]
        assert getattr(mock_cnx, register).call_count == count


def test_event_skipped(mock_libvirt):
    """
    Test that events are skipped if their ID isn't defined in the libvirt
    module (older libvirt)
    """
    mock_libvirt.mock_add_spec(
        [
            "openReadOnly",
            "virEventRegisterDefaultImpl",
            "virEventRunDefaultImpl",
            "VIR_DOMAIN_EVENT_ID_LIFECYCLE",
        ],
        spec_set=True,
    )

    libvirt_events.start("test:///", "test/prefix")

    # Check events registration and deregistration
    mock_cnx = mock_libvirt.openReadOnly.return_value

    mock_cnx.domainEventRegisterAny.assert_any_call(
        None,
        mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
        libvirt_events._domain_event_lifecycle_cb,
        {"prefix": "test/prefix", "object": "domain", "event": "lifecycle"},
    )

    # Network events should have been skipped
    mock_cnx.networkEventRegisterAny.assert_not_called()


def test_event_filtered(mock_libvirt):
    """
    Test that events are skipped if their ID isn't defined in the libvirt
    module (older libvirt)
    """
    libvirt_events.start("test", "test/prefix", "domain/lifecycle")

    # Check events registration and deregistration
    mock_cnx = mock_libvirt.openReadOnly.return_value

    mock_cnx.domainEventRegisterAny.assert_any_call(
        None,
        0,
        libvirt_events._domain_event_lifecycle_cb,
        {"prefix": "test/prefix", "object": "domain", "event": "lifecycle"},
    )

    # Network events should have been filtered out
    mock_cnx.networkEventRegisterAny.assert_not_called()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.rabbitmq_vhost as rabbitmq_vhost
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rabbitmq_vhost: {}}


# 'present' function tests: 1


def test_present():
    """
    Test to ensure the RabbitMQ VHost exists.
    """
    name = "virtual_host"

    ret = {
        "name": name,
        "changes": {"new": "virtual_host", "old": ""},
        "result": None,
        "comment": "Virtual Host 'virtual_host' will be created.",
    }

    mock = MagicMock(return_value=False)
    with patch.dict(rabbitmq_vhost.__salt__, {"rabbitmq.vhost_exists": mock}):
        with patch.dict(rabbitmq_vhost.__opts__, {"test": True}):
            assert rabbitmq_vhost.present(name) == ret


# 'absent' function tests: 1


def test_absent():
    """
    Test to ensure the named user is absent.
    """
    name = "myqueue"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Virtual Host '{}' is not present.".format(name),
    }

    mock = MagicMock(return_value=False)
    with patch.dict(rabbitmq_vhost.__salt__, {"rabbitmq.vhost_exists": mock}):
        assert rabbitmq_vhost.absent(name) == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.at as at
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {at: {}}


def test_present():
    """
    Test to add a job to queue.
    """
    name = "jboss"
    timespec = "9:09 11/04/15"
    tag = "love"
    user = "jam"

    # mock for at.at module call
    mock_atat = {
        "jobs": [
            {
                "date": "2015-11-04",
                "job": "1476031633.a",
                "queue": "a",
                "tag": tag,
                "time": "09:09:00",
                "user": user,
            },
        ],
    }

    # normale return
    ret = {
        "name": name,
        "result": True,
        "changes": {
            "date": "2015-11-04",
            "job": "1476031633.a",
            "queue": "a",
            "tag": "love",
            "time": "09:09:00",
            "user": "jam",
        },
        "comment": "job {name} added and will run on {timespec}".format(
            name=name, timespec=timespec
        ),
    }

    # unknown user return
    ret_user = {}
    ret_user.update(ret)
    ret_user.update(
        {
            "result": False,
            "changes": {},
            "comment": "user {} does not exists".format(user),
        }
    )

    # add a job
    mock = MagicMock(return_value=mock_atat)
    with patch.dict(at.__opts__, {"test": False}):
        with patch.dict(at.__salt__, {"at.at": mock}):
            assert at.present(name, timespec, tag) == ret

    # add a job with a non-existing user
    mock = MagicMock(return_value=False)
    with patch.dict(at.__opts__, {"test": False}):
        with patch.dict(at.__salt__, {"user.info": mock}):
            assert at.present(name, timespec, tag, user) == ret_user

    # add a job with test=True
    with patch.dict(at.__opts__, {"test": True}):
        ret_test = {}
        ret_test.update(ret)
        ret_test.update({"result": None, "changes": {}})
        assert at.present(name, timespec, tag, user) == ret_test


def test_absent():
    """
    Test to remove a job from queue
    """
    name = "jboss"
    tag = "rose"
    user = "jam"

    # mock for at.atrm module call
    mock_atatrm = {
        "jobs": {"removed": ["1476033859.a", "1476033855.a"], "tag": None},
    }

    # mock for at.jobcheck module call
    mock_atjobcheck = {
        "jobs": [
            {
                "date": "2015-11-04",
                "job": "1476031633.a",
                "queue": "a",
                "tag": tag,
                "time": "09:09:00",
                "user": user,
            },
        ],
    }

    # normal return
    ret = {
        "name": name,
        "result": True,
        "changes": {"removed": ["1476033859.a", "1476033855.a"]},
        "comment": "removed 2 job(s)",
    }

    # remove a job with test=True
    with patch.dict(at.__opts__, {"test": True}):
        ret_test = {}
        ret_test.update(ret)
        ret_test.update({"result": None, "changes": {}, "comment": "removed ? job(s)"})
        assert at.absent(name) == ret_test

    # remove a job and pass limit parameter
    with patch.dict(at.__opts__, {"test": False}):
        ret_limit = {}
        ret_limit.update(ret)
        ret_limit.update(
            {
                "result": False,
                "changes": {},
                "comment": "limit parameter not supported {}".format(name),
            }
        )
        assert at.absent(name, limit="all") == ret_limit

    # remove all jobs (2 jobs found)
    mock = MagicMock(return_value=mock_atatrm)
    with patch.dict(at.__salt__, {"at.atrm": mock}):
        with patch.dict(at.__opts__, {"test": False}):
            assert at.absent(name) == ret

    # remove all jobs (0 jobs found)
    mock_atatrm_nojobs = {}
    mock_atatrm_nojobs.update(mock_atatrm)
    mock_atatrm_nojobs.update({"jobs": {"removed": []}})
    mock = MagicMock(return_value=mock_atatrm_nojobs)
    with patch.dict(at.__salt__, {"at.atrm": mock}):
        with patch.dict(at.__opts__, {"test": False}):
            ret_nojobs = {}
            ret_nojobs.update(ret)
            ret_nojobs.update(
                {"changes": {}, "comment": ret["comment"].replace("2", "0")}
            )
            assert at.absent(name) == ret_nojobs

    # remove all tagged jobs (1 jobs found)
    mock_atatrm_tag = {}
    mock_atatrm_tag.update(mock_atatrm)
    mock_atatrm_tag.update({"jobs": {"removed": ["1476031633.a"], "tag": "rose"}})
    mock = MagicMock(return_value=mock_atatrm_tag)
    with patch.dict(at.__salt__, {"at.atrm": mock}):
        mock = MagicMock(return_value=mock_atjobcheck)
        with patch.dict(at.__salt__, {"at.jobcheck": mock}):
            with patch.dict(at.__opts__, {"test": False}):
                ret_tag = {}
                ret_tag.update(ret)
                ret_tag.update(
                    {
                        "changes": {"removed": ["1476031633.a"]},
                        "comment": ret["comment"].replace("2", "1"),
                    }
                )
                assert at.absent(name, tag=tag) == ret_tag



"""
    tests.support.cli_scripts
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Code to generate Salt CLI scripts for test runs
"""


import logging
import os

from saltfactories.utils.cli_scripts import generate_script

log = logging.getLogger(__name__)


def get_script_path(bin_dir, script_name):
    """
    Return the path to a testing runtime script, generating one if it does not yet exist
    """
    # Late import
    from tests.support.runtests import RUNTIME_VARS

    if not os.path.isdir(bin_dir):
        os.makedirs(bin_dir)

    cli_script_name = "cli_{}.py".format(script_name.replace("-", "_"))
    script_path = os.path.join(bin_dir, cli_script_name)

    if not os.path.isfile(script_path):
        kwargs = {
            "code_dir": str(RUNTIME_VARS.CODE_DIR),
            "bin_dir": bin_dir,
            "script_name": script_name,
            "coverage_rc_path": os.environ.get("COVERAGE_PROCESS_START"),
            "coverage_db_path": os.environ.get("COVERAGE_FILE"),
            "inject_sitecustomize": "COVERAGE_PROCESS_START" in os.environ,
        }
        generate_script(**kwargs)
    log.info("Returning script path %r", script_path)
    return script_path


class ScriptPathMixin:
    def get_script_path(self, script_name):
        """
        Return the path to a testing runtime script
        """
        # Late import
        from tests.support.runtests import RUNTIME_VARS

        return get_script_path(RUNTIME_VARS.TMP_SCRIPT_DIR, script_name)



"""
The networking module for Debian-based distros

References:

* http://www.debian.org/doc/manuals/debian-reference/ch05.en.html
"""
import functools
import io
import logging
import os
import os.path
import re
import time

import jinja2
import jinja2.exceptions

import salt.utils.dns
import salt.utils.files
import salt.utils.odict
import salt.utils.stringutils
import salt.utils.templates
import salt.utils.validate.net

log = logging.getLogger(__name__)

# Set up template environment
JINJA = jinja2.Environment(
    loader=jinja2.FileSystemLoader(
        os.path.join(salt.utils.templates.TEMPLATE_DIRNAME, "debian_ip")
    )
)

# Define the module's virtual name
__virtualname__ = "ip"


def __virtual__():
    """
    Confine this module to Debian-based distros
    """
    if __grains__["os_family"] == "Debian":
        return __virtualname__
    return (False, "The debian_ip module could not be loaded: unsupported OS family")


_ETHTOOL_CONFIG_OPTS = {
    "speed": "link-speed",
    "duplex": "link-duplex",
    "autoneg": "ethernet-autoneg",
    "ethernet-port": "ethernet-port",
    "wol": "ethernet-wol",
    "driver-message-level": "driver-message-level",
    "ethernet-pause-rx": "ethernet-pause-rx",
    "ethernet-pause-tx": "ethernet-pause-tx",
    "ethernet-pause-autoneg": "ethernet-pause-autoneg",
    "rx": "offload-rx",
    "tx": "offload-tx",
    "sg": "offload-sg",
    "tso": "offload-tso",
    "ufo": "offload-ufo",
    "gso": "offload-gso",
    "gro": "offload-gro",
    "lro": "offload-lro",
    "hardware-irq-coalesce-adaptive-rx": "hardware-irq-coalesce-adaptive-rx",
    "hardware-irq-coalesce-adaptive-tx": "hardware-irq-coalesce-adaptive-tx",
    "hardware-irq-coalesce-rx-usecs": "hardware-irq-coalesce-rx-usecs",
    "hardware-irq-coalesce-rx-frames": "hardware-irq-coalesce-rx-frames",
    "hardware-dma-ring-rx": "hardware-dma-ring-rx",
    "hardware-dma-ring-rx-mini": "hardware-dma-ring-rx-mini",
    "hardware-dma-ring-rx-jumbo": "hardware-dma-ring-rx-jumbo",
    "hardware-dma-ring-tx": "hardware-dma-ring-tx",
}

_REV_ETHTOOL_CONFIG_OPTS = {
    "link-speed": "speed",
    "link-duplex": "duplex",
    "ethernet-autoneg": "autoneg",
    "ethernet-port": "ethernet-port",
    "ethernet-wol": "wol",
    "driver-message-level": "driver-message-level",
    "ethernet-pause-rx": "ethernet-pause-rx",
    "ethernet-pause-tx": "ethernet-pause-tx",
    "ethernet-pause-autoneg": "ethernet-pause-autoneg",
    "offload-rx": "rx",
    "offload-tx": "tx",
    "offload-sg": "sg",
    "offload-tso": "tso",
    "offload-ufo": "ufo",
    "offload-gso": "gso",
    "offload-lro": "lro",
    "offload-gro": "gro",
    "hardware-irq-coalesce-adaptive-rx": "hardware-irq-coalesce-adaptive-rx",
    "hardware-irq-coalesce-adaptive-tx": "hardware-irq-coalesce-adaptive-tx",
    "hardware-irq-coalesce-rx-usecs": "hardware-irq-coalesce-rx-usecs",
    "hardware-irq-coalesce-rx-frames": "hardware-irq-coalesce-rx-frames",
    "hardware-dma-ring-rx": "hardware-dma-ring-rx",
    "hardware-dma-ring-rx-mini": "hardware-dma-ring-rx-mini",
    "hardware-dma-ring-rx-jumbo": "hardware-dma-ring-rx-jumbo",
    "hardware-dma-ring-tx": "hardware-dma-ring-tx",
}

_DEB_CONFIG_PPPOE_OPTS = {
    "user": "user",
    "password": "password",
    "provider": "provider",
    "pppoe_iface": "pppoe_iface",
    "noipdefault": "noipdefault",
    "usepeerdns": "usepeerdns",
    "defaultroute": "defaultroute",
    "holdoff": "holdoff",
    "maxfail": "maxfail",
    "hide-password": "hide-password",
    "lcp-echo-interval": "lcp-echo-interval",
    "lcp-echo-failure": "lcp-echo-failure",
    "connect": "connect",
    "noauth": "noauth",
    "persist": "persist",
    "mtu": "mtu",
    "noaccomp": "noaccomp",
    "linkname": "linkname",
}

_DEB_ROUTES_FILE = "/etc/network/routes"
_DEB_NETWORK_FILE = "/etc/network/interfaces"
_DEB_NETWORK_DIR = "/etc/network/interfaces.d/"
_DEB_NETWORK_UP_DIR = "/etc/network/if-up.d/"
_DEB_NETWORK_DOWN_DIR = "/etc/network/if-down.d/"
_DEB_NETWORK_CONF_FILES = "/etc/modprobe.d/"
_DEB_NETWORKING_FILE = "/etc/default/networking"
_DEB_HOSTNAME_FILE = "/etc/hostname"
_DEB_RESOLV_FILE = "/etc/resolv.conf"
_DEB_PPP_DIR = "/etc/ppp/peers/"

_CONFIG_TRUE = ["yes", "on", "true", "1", True]
_CONFIG_FALSE = ["no", "off", "false", "0", False]
_IFACE_TYPES = [
    "eth",
    "bond",
    "alias",
    "clone",
    "ipsec",
    "dialup",
    "bridge",
    "slave",
    "vlan",
    "pppoe",
    "source",
]


def _error_msg_iface(iface, option, expected):
    """
    Build an appropriate error message from a given option and
    a list of expected values.
    """
    msg = "Invalid option -- Interface: {0}, Option: {1}, Expected: [{2}]"
    return msg.format(iface, option, "|".join(str(e) for e in expected))


def _error_msg_routes(iface, option, expected):
    """
    Build an appropriate error message from a given option and
    a list of expected values.
    """
    msg = "Invalid option -- Route interface: {0}, Option: {1}, Expected: [{2}]"
    return msg.format(iface, option, expected)


def _log_default_iface(iface, opt, value):
    log.info(
        "Using default option -- Interface: %s Option: %s Value: %s", iface, opt, value
    )


def _error_msg_network(option, expected):
    """
    Build an appropriate error message from a given option and
    a list of expected values.
    """
    msg = "Invalid network setting -- Setting: {0}, Expected: [{1}]"
    return msg.format(option, "|".join(str(e) for e in expected))


def _log_default_network(opt, value):
    log.info("Using existing setting -- Setting: %s Value: %s", opt, value)


def _raise_error_iface(iface, option, expected):
    """
    Log and raise an error with a logical formatted message.
    """
    msg = _error_msg_iface(iface, option, expected)
    log.error(msg)
    raise AttributeError(msg)


def _raise_error_network(option, expected):
    """
    Log and raise an error with a logical formatted message.
    """
    msg = _error_msg_network(option, expected)
    log.error(msg)
    raise AttributeError(msg)


def _raise_error_routes(iface, option, expected):
    """
    Log and raise an error with a logical formatted message.
    """
    msg = _error_msg_routes(iface, option, expected)
    log.error(msg)
    raise AttributeError(msg)


def _read_file(path):
    """
    Reads and returns the contents of a text file
    """
    try:
        with salt.utils.files.flopen(path, "rb") as contents:
            return [
                salt.utils.stringutils.to_str(line) for line in contents.readlines()
            ]
    except OSError:
        return ""


def _parse_resolve():
    """
    Parse /etc/resolv.conf
    """
    return salt.utils.dns.parse_resolv(_DEB_RESOLV_FILE)


def _parse_domainname():
    """
    Parse /etc/resolv.conf and return domainname
    """
    return _parse_resolve().get("domain", "")


def _parse_searchdomain():
    """
    Parse /etc/resolv.conf and return searchdomain
    """
    return _parse_resolve().get("search", "")


def _parse_hostname():
    """
    Parse /etc/hostname and return hostname
    """
    contents = _read_file(_DEB_HOSTNAME_FILE)
    if contents:
        return contents[0].split("\n")[0]
    else:
        return ""


def _parse_current_network_settings():
    """
    Parse /etc/default/networking and return current configuration
    """
    opts = salt.utils.odict.OrderedDict()
    opts["networking"] = ""

    if os.path.isfile(_DEB_NETWORKING_FILE):
        with salt.utils.files.fopen(_DEB_NETWORKING_FILE) as contents:
            for line in contents:
                salt.utils.stringutils.to_unicode(line)
                if line.startswith("#"):
                    continue
                elif line.startswith("CONFIGURE_INTERFACES"):
                    opts["networking"] = line.split("=", 1)[1].strip()

    hostname = _parse_hostname()
    domainname = _parse_domainname()
    searchdomain = _parse_searchdomain()

    opts["hostname"] = hostname
    opts["domainname"] = domainname
    opts["searchdomain"] = searchdomain
    return opts


# def __validator_func(value):
#    return (valid: True/False, (transformed) value, error message)


def __ipv4_quad(value):
    """validate an IPv4 address"""
    return (salt.utils.validate.net.ipv4_addr(value), value, "dotted IPv4 address")


def __ipv6(value):
    """validate an IPv6 address"""
    return (salt.utils.validate.net.ipv6_addr(value), value, "IPv6 address")


def __mac(value):
    """validate a mac address"""
    return (salt.utils.validate.net.mac(value), value, "MAC address")


def __anything(value):
    return (True, value, None)


def __int(value):
    """validate an integer"""
    valid, _value = False, value
    try:
        _value = int(value)
        valid = True
    except ValueError:
        pass
    return (valid, _value, "integer")


def __float(value):
    """validate a float"""
    valid, _value = False, value
    try:
        _value = float(value)
        valid = True
    except ValueError:
        pass
    return (valid, _value, "float")


def __ipv4_netmask(value):
    """validate an IPv4 dotted quad or integer CIDR netmask"""
    valid, errmsg = False, "dotted quad or integer CIDR (0->32)"
    valid, value, _ = __int(value)
    if not (valid and 0 <= value <= 32):
        valid = salt.utils.validate.net.netmask(value)
    return (valid, value, errmsg)


def __ipv6_netmask(value):
    """validate an IPv6 integer netmask"""
    valid, errmsg = False, "IPv6 netmask (0->128)"
    valid, value, _ = __int(value)
    valid = valid and 0 <= value <= 128
    return (valid, value, errmsg)


def __within2(value, within=None, errmsg=None, dtype=None):
    """validate that a value is in ``within`` and optionally a ``dtype``"""
    valid, _value = False, value
    if dtype:
        try:
            _value = dtype(value)  # TODO: this is a bit loose when dtype is a class
            valid = _value in within
        except ValueError:
            pass
    else:
        valid = _value in within
    if errmsg is None:
        if dtype:
            typename = getattr(
                dtype,
                "__name__",
                hasattr(dtype, "__class__") and getattr(dtype.__class__, "name", dtype),
            )
            errmsg = "{} within '{}'".format(typename, within)
        else:
            errmsg = "within '{}'".format(within)
    return (valid, _value, errmsg)


def __within(within=None, errmsg=None, dtype=None):
    return functools.partial(__within2, within=within, errmsg=errmsg, dtype=dtype)


def __space_delimited_list(value):
    """validate that a value contains one or more space-delimited values"""
    if isinstance(value, str):
        value = value.strip().split()

    if hasattr(value, "__iter__") and value != []:
        return (True, value, "space-delimited string")
    else:
        return (
            False,
            value,
            "{} is not a valid space-delimited value.\n".format(value),
        )


SALT_ATTR_TO_DEBIAN_ATTR_MAP = {
    "dns": "dns-nameservers",
    "search": "dns-search",
    "hwaddr": "hwaddress",  # TODO: this limits bootp functionality
    "ipaddr": "address",
    "ipaddrs": "addresses",
}


DEBIAN_ATTR_TO_SALT_ATTR_MAP = {v: k for (k, v) in SALT_ATTR_TO_DEBIAN_ATTR_MAP.items()}

# TODO
DEBIAN_ATTR_TO_SALT_ATTR_MAP["address"] = "address"
DEBIAN_ATTR_TO_SALT_ATTR_MAP["hwaddress"] = "hwaddress"

IPV4_VALID_PROTO = ["bootp", "dhcp", "static", "manual", "loopback", "ppp"]

IPV4_ATTR_MAP = {
    "proto": __within(IPV4_VALID_PROTO, dtype=str),
    # ipv4 static & manual
    "address": __ipv4_quad,
    "addresses": __anything,
    "netmask": __ipv4_netmask,
    "broadcast": __ipv4_quad,
    "metric": __int,
    "gateway": __ipv4_quad,  # supports a colon-delimited list
    "pointopoint": __ipv4_quad,
    "hwaddress": __mac,
    "mtu": __int,
    "scope": __within(["global", "link", "host"], dtype=str),
    # dhcp
    "hostname": __anything,
    "leasehours": __int,
    "leasetime": __int,
    "vendor": __anything,
    "client": __anything,
    # bootp
    "bootfile": __anything,
    "server": __ipv4_quad,
    "hwaddr": __mac,
    # tunnel
    "mode": __within(["gre", "GRE", "ipip", "IPIP", "802.3ad"], dtype=str),
    "endpoint": __ipv4_quad,
    "dstaddr": __ipv4_quad,
    "local": __ipv4_quad,
    "ttl": __int,
    # bond
    "slaves": __anything,
    # ppp
    "provider": __anything,
    "unit": __int,
    "options": __anything,
    # resolvconf
    "dns-nameservers": __space_delimited_list,
    "dns-search": __space_delimited_list,
    #
    "vlan-raw-device": __anything,
    #
    "network": __anything,  # i don't know what this is
    "test": __anything,  # TODO
    "enable_ipv4": __anything,  # TODO
    "enable_ipv6": __anything,  # TODO
}


IPV6_VALID_PROTO = ["auto", "loopback", "static", "manual", "dhcp", "v4tunnel", "6to4"]

IPV6_ATTR_MAP = {
    "proto": __within(IPV6_VALID_PROTO),
    # ipv6 static & manual
    "address": __ipv6,
    "addresses": __anything,
    "netmask": __ipv6_netmask,
    "broadcast": __ipv6,
    "gateway": __ipv6,  # supports a colon-delimited list
    "hwaddress": __mac,
    "mtu": __int,
    "scope": __within(["global", "site", "link", "host"], dtype=str),
    # inet6 auto
    "privext": __within([0, 1, 2], dtype=int),
    "dhcp": __within([0, 1], dtype=int),
    # inet6 static & manual & dhcp
    "media": __anything,
    "accept_ra": __within([0, 1, 2], dtype=int),
    "autoconf": __within([0, 1], dtype=int),
    "preferred-lifetime": __int,
    "dad-attempts": __int,  # 0 to disable
    "dad-interval": __float,
    # bond
    "slaves": __anything,
    # tunnel
    "mode": __within(["gre", "GRE", "ipip", "IPIP", "802.3ad"], dtype=str),
    "endpoint": __ipv4_quad,
    "local": __ipv4_quad,
    "ttl": __int,
    # resolvconf
    "dns-nameservers": __space_delimited_list,
    "dns-search": __space_delimited_list,
    #
    "vlan-raw-device": __anything,
    "test": __anything,  # TODO
    "enable_ipv4": __anything,  # TODO
    "enable_ipv6": __anything,  # TODO
}


WIRELESS_ATTR_MAP = {
    "wireless-essid": __anything,
    "wireless-mode": __anything,  # TODO
    "wpa-ap-scan": __within([0, 1, 2], dtype=int),  # TODO
    "wpa-conf": __anything,
    "wpa-driver": __anything,
    "wpa-group": __anything,
    "wpa-key-mgmt": __anything,
    "wpa-pairwise": __anything,
    "wpa-psk": __anything,
    "wpa-proto": __anything,  # partial(__within,
    "wpa-roam": __anything,
    "wpa-ssid": __anything,  # TODO
}

ATTRMAPS = {
    "inet": [IPV4_ATTR_MAP, WIRELESS_ATTR_MAP],
    "inet6": [IPV6_ATTR_MAP, WIRELESS_ATTR_MAP],
}


def _validate_interface_option(attr, value, addrfam="inet"):
    """lookup the validation function for a [addrfam][attr] and
    return the results

    :param attr: attribute name
    :param value: raw setting value
    :param addrfam: address family (inet, inet6,
    """
    valid, _value, errmsg = False, value, "Unknown validator"
    attrmaps = ATTRMAPS.get(addrfam, [])
    for attrmap in attrmaps:
        if attr in attrmap:
            validate_func = attrmap[attr]
            (valid, _value, errmsg) = validate_func(value)
            break
    return (valid, _value, errmsg)


def _attrmaps_contain_attr(attr):
    return attr in WIRELESS_ATTR_MAP or attr in IPV4_ATTR_MAP or attr in IPV6_ATTR_MAP


def _parse_interfaces(interface_files=None):
    """
    Parse /etc/network/interfaces and return current configured interfaces
    """
    if interface_files is None:
        interface_files = []
        # Add this later.
        if os.path.exists(_DEB_NETWORK_DIR):
            interface_files += [
                "{}/{}".format(_DEB_NETWORK_DIR, dir)
                for dir in os.listdir(_DEB_NETWORK_DIR)
            ]

        if os.path.isfile(_DEB_NETWORK_FILE):
            interface_files.insert(0, _DEB_NETWORK_FILE)

    adapters = salt.utils.odict.OrderedDict()
    method = -1

    for interface_file in interface_files:
        with salt.utils.files.fopen(interface_file) as interfaces:
            # This ensures iface_dict exists, but does not ensure we're not reading a new interface.
            iface_dict = {}
            for line in interfaces:
                line = salt.utils.stringutils.to_unicode(line)
                # Identify the clauses by the first word of each line.
                # Go to the next line if the current line is a comment
                # or all spaces.
                if line.lstrip().startswith("#") or line.isspace():
                    continue
                # Parse the iface clause
                if line.startswith("iface"):
                    sline = line.split()

                    if len(sline) != 4:
                        msg = "Interface file malformed: {0}."
                        msg = msg.format(sline)
                        log.error(msg)
                        raise AttributeError(msg)

                    iface_name = sline[1]
                    addrfam = sline[2]
                    method = sline[3]

                    # Create item in dict, if not already there
                    if iface_name not in adapters:
                        adapters[iface_name] = salt.utils.odict.OrderedDict()

                    # Create item in dict, if not already there
                    if "data" not in adapters[iface_name]:
                        adapters[iface_name]["data"] = salt.utils.odict.OrderedDict()

                    if addrfam not in adapters[iface_name]["data"]:
                        adapters[iface_name]["data"][
                            addrfam
                        ] = salt.utils.odict.OrderedDict()

                    iface_dict = adapters[iface_name]["data"][addrfam]

                    iface_dict["addrfam"] = addrfam
                    iface_dict["proto"] = method
                    iface_dict["filename"] = interface_file

                # Parse the detail clauses.
                elif line[0].isspace():
                    sline = line.split()

                    # conf file attr: dns-nameservers
                    # salt states.network attr: dns

                    attr, valuestr = line.rstrip().split(None, 1)
                    if _attrmaps_contain_attr(attr):
                        if "-" in attr:
                            attrname = attr.replace("-", "_")
                        else:
                            attrname = attr
                        (valid, value, errmsg) = _validate_interface_option(
                            attr, valuestr, addrfam
                        )
                        if attrname == "address" and "address" in iface_dict:
                            if "addresses" not in iface_dict:
                                iface_dict["addresses"] = []
                            iface_dict["addresses"].append(value)
                        else:
                            iface_dict[attrname] = value

                    elif attr in _REV_ETHTOOL_CONFIG_OPTS:
                        if "ethtool" not in iface_dict:
                            iface_dict["ethtool"] = salt.utils.odict.OrderedDict()
                        iface_dict["ethtool"][attr] = valuestr

                    elif attr.startswith("bond"):
                        opt = re.split(r"[_-]", attr, maxsplit=1)[1]
                        if "bonding" not in iface_dict:
                            iface_dict["bonding"] = salt.utils.odict.OrderedDict()
                        iface_dict["bonding"][opt] = valuestr

                    elif attr.startswith("bridge"):
                        opt = re.split(r"[_-]", attr, maxsplit=1)[1]
                        if "bridging" not in iface_dict:
                            iface_dict["bridging"] = salt.utils.odict.OrderedDict()
                        iface_dict["bridging"][opt] = valuestr

                    elif attr in [
                        "up",
                        "pre-up",
                        "post-up",
                        "down",
                        "pre-down",
                        "post-down",
                    ]:
                        cmd = valuestr
                        cmd_key = "{}_cmds".format(re.sub("-", "_", attr))
                        if cmd_key not in iface_dict:
                            iface_dict[cmd_key] = []
                        iface_dict[cmd_key].append(cmd)

                elif line.startswith("auto"):
                    for word in line.split()[1:]:
                        if word not in adapters:
                            adapters[word] = salt.utils.odict.OrderedDict()
                        adapters[word]["enabled"] = True

                elif line.startswith("allow-hotplug"):
                    for word in line.split()[1:]:
                        if word not in adapters:
                            adapters[word] = salt.utils.odict.OrderedDict()
                        adapters[word]["hotplug"] = True

                elif line.startswith("source"):
                    if "source" not in adapters:
                        adapters["source"] = salt.utils.odict.OrderedDict()

                    # Create item in dict, if not already there
                    if "data" not in adapters["source"]:
                        adapters["source"]["data"] = salt.utils.odict.OrderedDict()
                        adapters["source"]["data"]["sources"] = []
                    adapters["source"]["data"]["sources"].append(line.split()[1])

    adapters = _filter_malformed_interfaces(adapters=adapters)
    return adapters


def _filter_malformed_interfaces(*, adapters):
    # Return a sorted list of the keys for bond, bridge and ethtool options to
    # ensure a consistent order
    for iface_name in list(adapters):
        if iface_name == "source":
            continue
        if "data" not in adapters[iface_name]:
            msg = "Interface file malformed for interface: {}.".format(iface_name)
            log.error(msg)
            adapters.pop(iface_name)
            continue
        for opt in ["ethtool", "bonding", "bridging"]:
            for inet in ["inet", "inet6"]:
                if inet in adapters[iface_name]["data"]:
                    if opt in adapters[iface_name]["data"][inet]:
                        opt_keys = sorted(
                            adapters[iface_name]["data"][inet][opt].keys()
                        )
                        adapters[iface_name]["data"][inet][opt + "_keys"] = opt_keys
    return adapters


def _parse_ethtool_opts(opts, iface):
    """
    Filters given options and outputs valid settings for ETHTOOLS_OPTS
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    config = {}

    if "autoneg" in opts:
        if opts["autoneg"] in _CONFIG_TRUE:
            config.update({"autoneg": "on"})
        elif opts["autoneg"] in _CONFIG_FALSE:
            config.update({"autoneg": "off"})
        else:
            _raise_error_iface(iface, "autoneg", _CONFIG_TRUE + _CONFIG_FALSE)

    if "duplex" in opts:
        valid = ["full", "half"]
        if opts["duplex"] in valid:
            config.update({"duplex": opts["duplex"]})
        else:
            _raise_error_iface(iface, "duplex", valid)

    if "speed" in opts:
        valid = ["10", "100", "1000", "10000"]
        if str(opts["speed"]) in valid:
            config.update({"speed": opts["speed"]})
        else:
            _raise_error_iface(iface, opts["speed"], valid)

    valid = _CONFIG_TRUE + _CONFIG_FALSE
    for option in ("rx", "tx", "sg", "tso", "ufo", "gso", "gro", "lro"):
        if option in opts:
            if opts[option] in _CONFIG_TRUE:
                config.update({option: "on"})
            elif opts[option] in _CONFIG_FALSE:
                config.update({option: "off"})
            else:
                _raise_error_iface(iface, option, valid)

    return config


def _parse_ethtool_pppoe_opts(opts, iface):
    """
    Filters given options and outputs valid settings for ETHTOOLS_PPPOE_OPTS
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    config = {}

    for opt in _DEB_CONFIG_PPPOE_OPTS:
        if opt in opts:
            config[opt] = opts[opt]

    if "provider" in opts and not opts["provider"]:
        _raise_error_iface(iface, "provider", _CONFIG_TRUE + _CONFIG_FALSE)

    valid = _CONFIG_TRUE + _CONFIG_FALSE
    for option in (
        "noipdefault",
        "usepeerdns",
        "defaultroute",
        "hide-password",
        "noauth",
        "persist",
        "noaccomp",
    ):
        if option in opts:
            if opts[option] in _CONFIG_TRUE:
                config.update({option: "True"})
            elif opts[option] in _CONFIG_FALSE:
                config.update({option: "False"})
            else:
                _raise_error_iface(iface, option, valid)

    return config


def _parse_settings_bond(opts, iface):
    """
    Filters given options and outputs valid settings for requested
    operation. If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """

    bond_def = {
        # 803.ad aggregation selection logic
        # 0 for stable (default)
        # 1 for bandwidth
        # 2 for count
        "ad_select": "0",
        # Max number of transmit queues (default = 16)
        "tx_queues": "16",
        # Link monitoring in milliseconds. Most NICs support this
        "miimon": "100",
        # ARP interval in milliseconds
        "arp_interval": "250",
        # Delay before considering link down in milliseconds (miimon * 2)
        "downdelay": "200",
        # lacp_rate 0: Slow - every 30 seconds
        # lacp_rate 1: Fast - every 1 second
        "lacp_rate": "0",
        # Max bonds for this driver
        "max_bonds": "1",
        # Specifies the time, in milliseconds, to wait before
        # enabling a slave after a link recovery has been
        # detected. Only used with miimon.
        "updelay": "0",
        # Used with miimon.
        # On: driver sends mii
        # Off: ethtool sends mii
        "use_carrier": "on",
        # Default. Don't change unless you know what you are doing.
        "xmit_hash_policy": "layer2",
    }

    if opts["mode"] in ["balance-rr", "0"]:
        log.info("Device: %s Bonding Mode: load balancing (round-robin)", iface)
        return _parse_settings_bond_0(opts, iface, bond_def)
    elif opts["mode"] in ["active-backup", "1"]:
        log.info("Device: %s Bonding Mode: fault-tolerance (active-backup)", iface)
        return _parse_settings_bond_1(opts, iface, bond_def)
    elif opts["mode"] in ["balance-xor", "2"]:
        log.info("Device: %s Bonding Mode: load balancing (xor)", iface)
        return _parse_settings_bond_2(opts, iface, bond_def)
    elif opts["mode"] in ["broadcast", "3"]:
        log.info("Device: %s Bonding Mode: fault-tolerance (broadcast)", iface)
        return _parse_settings_bond_3(opts, iface, bond_def)
    elif opts["mode"] in ["802.3ad", "4"]:
        log.info(
            "Device: %s Bonding Mode: IEEE 802.3ad Dynamic link aggregation", iface
        )
        return _parse_settings_bond_4(opts, iface, bond_def)
    elif opts["mode"] in ["balance-tlb", "5"]:
        log.info("Device: %s Bonding Mode: transmit load balancing", iface)
        return _parse_settings_bond_5(opts, iface, bond_def)
    elif opts["mode"] in ["balance-alb", "6"]:
        log.info("Device: %s Bonding Mode: adaptive load balancing", iface)
        return _parse_settings_bond_6(opts, iface, bond_def)
    else:
        valid = [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "balance-rr",
            "active-backup",
            "balance-xor",
            "broadcast",
            "802.3ad",
            "balance-tlb",
            "balance-alb",
        ]
        _raise_error_iface(iface, "mode", valid)


def _parse_settings_bond_0(opts, iface, bond_def):
    """
    Filters given options and outputs valid settings for bond0.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    bond = {"mode": "0"}

    # ARP targets in n.n.n.n form
    valid = ["list of ips (up to 16)"]
    if "arp_ip_target" in opts:
        if isinstance(opts["arp_ip_target"], list):
            if 1 <= len(opts["arp_ip_target"]) <= 16:
                bond.update({"arp_ip_target": ""})
                for ip in opts["arp_ip_target"]:  # pylint: disable=C0103
                    if len(bond["arp_ip_target"]) > 0:
                        bond["arp_ip_target"] = bond["arp_ip_target"] + "," + ip
                    else:
                        bond["arp_ip_target"] = ip
            else:
                _raise_error_iface(iface, "arp_ip_target", valid)
        else:
            _raise_error_iface(iface, "arp_ip_target", valid)
    else:
        _raise_error_iface(iface, "arp_ip_target", valid)

    if "arp_interval" in opts:
        try:
            int(opts["arp_interval"])
            bond.update({"arp_interval": opts["arp_interval"]})
        except ValueError:
            _raise_error_iface(iface, "arp_interval", ["integer"])
    else:
        _log_default_iface(iface, "arp_interval", bond_def["arp_interval"])
        bond.update({"arp_interval": bond_def["arp_interval"]})

    return bond


def _parse_settings_bond_1(opts, iface, bond_def):

    """
    Filters given options and outputs valid settings for bond1.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    bond = {"mode": "1"}

    for binding in ["miimon", "downdelay", "updelay"]:
        if binding in opts:
            try:
                int(opts[binding])
                bond.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, ["integer"])
        else:
            _log_default_iface(iface, binding, bond_def[binding])
            bond.update({binding: bond_def[binding]})

    if "primary" in opts:
        bond.update({"primary": opts["primary"]})

    if not (__grains__["os"] == "Ubuntu" and __grains__["osrelease_info"][0] >= 16):
        if "use_carrier" in opts:
            if opts["use_carrier"] in _CONFIG_TRUE:
                bond.update({"use_carrier": "1"})
            elif opts["use_carrier"] in _CONFIG_FALSE:
                bond.update({"use_carrier": "0"})
            else:
                valid = _CONFIG_TRUE + _CONFIG_FALSE
                _raise_error_iface(iface, "use_carrier", valid)
        else:
            _log_default_iface(iface, "use_carrier", bond_def["use_carrier"])
            bond.update({"use_carrier": bond_def["use_carrier"]})

    return bond


def _parse_settings_bond_2(opts, iface, bond_def):
    """
    Filters given options and outputs valid settings for bond2.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """

    bond = {"mode": "2"}

    valid = ["list of ips (up to 16)"]
    if "arp_ip_target" in opts:
        if isinstance(opts["arp_ip_target"], list):
            if 1 <= len(opts["arp_ip_target"]) <= 16:
                bond.update({"arp_ip_target": ""})
                for ip in opts["arp_ip_target"]:  # pylint: disable=C0103
                    if len(bond["arp_ip_target"]) > 0:
                        bond["arp_ip_target"] = bond["arp_ip_target"] + "," + ip
                    else:
                        bond["arp_ip_target"] = ip
            else:
                _raise_error_iface(iface, "arp_ip_target", valid)
        else:
            _raise_error_iface(iface, "arp_ip_target", valid)
    else:
        _raise_error_iface(iface, "arp_ip_target", valid)

    if "arp_interval" in opts:
        try:
            int(opts["arp_interval"])
            bond.update({"arp_interval": opts["arp_interval"]})
        except ValueError:
            _raise_error_iface(iface, "arp_interval", ["integer"])
    else:
        _log_default_iface(iface, "arp_interval", bond_def["arp_interval"])
        bond.update({"arp_interval": bond_def["arp_interval"]})

    if "hashing-algorithm" in opts:
        valid = ["layer2", "layer2+3", "layer3+4"]
        if opts["hashing-algorithm"] in valid:
            bond.update({"xmit_hash_policy": opts["hashing-algorithm"]})
        else:
            _raise_error_iface(iface, "hashing-algorithm", valid)

    return bond


def _parse_settings_bond_3(opts, iface, bond_def):

    """
    Filters given options and outputs valid settings for bond3.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    bond = {"mode": "3"}

    for binding in ["miimon", "downdelay", "updelay"]:
        if binding in opts:
            try:
                int(opts[binding])
                bond.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, ["integer"])
        else:
            _log_default_iface(iface, binding, bond_def[binding])
            bond.update({binding: bond_def[binding]})

    if "use_carrier" in opts:
        if opts["use_carrier"] in _CONFIG_TRUE:
            bond.update({"use_carrier": "1"})
        elif opts["use_carrier"] in _CONFIG_FALSE:
            bond.update({"use_carrier": "0"})
        else:
            valid = _CONFIG_TRUE + _CONFIG_FALSE
            _raise_error_iface(iface, "use_carrier", valid)
    else:
        _log_default_iface(iface, "use_carrier", bond_def["use_carrier"])
        bond.update({"use_carrier": bond_def["use_carrier"]})

    return bond


def _parse_settings_bond_4(opts, iface, bond_def):
    """
    Filters given options and outputs valid settings for bond4.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """

    bond = {"mode": "4"}

    for binding in ["miimon", "downdelay", "updelay", "lacp_rate", "ad_select"]:
        if binding in opts:
            if binding == "lacp_rate":
                if opts[binding] == "fast":
                    opts.update({binding: "1"})
                if opts[binding] == "slow":
                    opts.update({binding: "0"})
                valid = ["fast", "1", "slow", "0"]
            else:
                valid = ["integer"]
            try:
                int(opts[binding])
                bond.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, valid)
        else:
            _log_default_iface(iface, binding, bond_def[binding])
            bond.update({binding: bond_def[binding]})

    if "use_carrier" in opts:
        if opts["use_carrier"] in _CONFIG_TRUE:
            bond.update({"use_carrier": "1"})
        elif opts["use_carrier"] in _CONFIG_FALSE:
            bond.update({"use_carrier": "0"})
        else:
            valid = _CONFIG_TRUE + _CONFIG_FALSE
            _raise_error_iface(iface, "use_carrier", valid)
    else:
        _log_default_iface(iface, "use_carrier", bond_def["use_carrier"])
        bond.update({"use_carrier": bond_def["use_carrier"]})

    if "hashing-algorithm" in opts:
        valid = ["layer2", "layer2+3", "layer3+4"]
        if opts["hashing-algorithm"] in valid:
            bond.update({"xmit_hash_policy": opts["hashing-algorithm"]})
        else:
            _raise_error_iface(iface, "hashing-algorithm", valid)

    return bond


def _parse_settings_bond_5(opts, iface, bond_def):

    """
    Filters given options and outputs valid settings for bond5.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    bond = {"mode": "5"}

    for binding in ["miimon", "downdelay", "updelay"]:
        if binding in opts:
            try:
                int(opts[binding])
                bond.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, ["integer"])
        else:
            _log_default_iface(iface, binding, bond_def[binding])
            bond.update({binding: bond_def[binding]})

    if "use_carrier" in opts:
        if opts["use_carrier"] in _CONFIG_TRUE:
            bond.update({"use_carrier": "1"})
        elif opts["use_carrier"] in _CONFIG_FALSE:
            bond.update({"use_carrier": "0"})
        else:
            valid = _CONFIG_TRUE + _CONFIG_FALSE
            _raise_error_iface(iface, "use_carrier", valid)
    else:
        _log_default_iface(iface, "use_carrier", bond_def["use_carrier"])
        bond.update({"use_carrier": bond_def["use_carrier"]})

    if "primary" in opts:
        bond.update({"primary": opts["primary"]})

    return bond


def _parse_settings_bond_6(opts, iface, bond_def):

    """
    Filters given options and outputs valid settings for bond6.
    If an option has a value that is not expected, this
    function will log what the Interface, Setting and what it was
    expecting.
    """
    bond = {"mode": "6"}

    for binding in ["miimon", "downdelay", "updelay"]:
        if binding in opts:
            try:
                int(opts[binding])
                bond.update({binding: opts[binding]})
            except ValueError:
                _raise_error_iface(iface, binding, ["integer"])
        else:
            _log_default_iface(iface, binding, bond_def[binding])
            bond.update({binding: bond_def[binding]})

    if "use_carrier" in opts:
        if opts["use_carrier"] in _CONFIG_TRUE:
            bond.update({"use_carrier": "1"})
        elif opts["use_carrier"] in _CONFIG_FALSE:
            bond.update({"use_carrier": "0"})
        else:
            valid = _CONFIG_TRUE + _CONFIG_FALSE
            _raise_error_iface(iface, "use_carrier", valid)
    else:
        _log_default_iface(iface, "use_carrier", bond_def["use_carrier"])
        bond.update({"use_carrier": bond_def["use_carrier"]})

    if "primary" in opts:
        bond.update({"primary": opts["primary"]})

    return bond


def _parse_bridge_opts(opts, iface):
    """
    Filters given options and outputs valid settings for BRIDGING_OPTS
    If an option has a value that is not expected, this
    function will log the Interface, Setting and what was expected.
    """
    config = {}

    if "ports" in opts:
        if isinstance(opts["ports"], list):
            opts["ports"] = " ".join(opts["ports"])
        config.update({"ports": opts["ports"]})

    for opt in ["ageing", "fd", "gcint", "hello", "maxage"]:
        if opt in opts:
            try:
                float(opts[opt])
                config.update({opt: opts[opt]})
            except ValueError:
                _raise_error_iface(iface, opt, ["float"])

    for opt in ["bridgeprio", "maxwait"]:
        if opt in opts:
            if isinstance(opts[opt], int):
                config.update({opt: opts[opt]})
            else:
                _raise_error_iface(iface, opt, ["integer"])

    if "hw" in opts:
        # match 12 hex digits with either : or - as separators between pairs
        if re.match(
            "[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", opts["hw"].lower()
        ):
            config.update({"hw": opts["hw"]})
        else:
            _raise_error_iface(iface, "hw", ["valid MAC address"])

    for opt in ["pathcost", "portprio"]:
        if opt in opts:
            try:
                port, cost_or_prio = opts[opt].split()
                int(cost_or_prio)
                config.update({opt: "{} {}".format(port, cost_or_prio)})
            except ValueError:
                _raise_error_iface(iface, opt, ["interface integer"])

    if "stp" in opts:
        if opts["stp"] in _CONFIG_TRUE:
            config.update({"stp": "on"})
        elif opts["stp"] in _CONFIG_FALSE:
            config.update({"stp": "off"})
        else:
            _raise_error_iface(iface, "stp", _CONFIG_TRUE + _CONFIG_FALSE)

    if "waitport" in opts:
        if isinstance(opts["waitport"], int):
            config.update({"waitport": opts["waitport"]})
        else:
            values = opts["waitport"].split()
            waitport_time = values.pop(0)
            if waitport_time.isdigit() and values:
                config.update(
                    {"waitport": "{} {}".format(waitport_time, " ".join(values))}
                )
            else:
                _raise_error_iface(iface, opt, ["integer [interfaces]"])

    return config


def _parse_settings_eth(opts, iface_type, enabled, iface):
    """
    Filters given options and outputs valid settings for a
    network interface.
    """
    adapters = salt.utils.odict.OrderedDict()
    adapters[iface] = salt.utils.odict.OrderedDict()

    adapters[iface]["type"] = iface_type

    adapters[iface]["data"] = salt.utils.odict.OrderedDict()
    iface_data = adapters[iface]["data"]
    iface_data["inet"] = salt.utils.odict.OrderedDict()
    iface_data["inet6"] = salt.utils.odict.OrderedDict()

    if enabled:
        adapters[iface]["enabled"] = True

    if opts.get("hotplug", False):
        adapters[iface]["hotplug"] = True

    if opts.get("enable_ipv6", None) and opts.get("iface_type", "") == "vlan":
        iface_data["inet6"]["vlan_raw_device"] = re.sub(r"\.\d*", "", iface)

    for addrfam in ["inet", "inet6"]:
        if iface_type not in ["bridge"]:
            tmp_ethtool = _parse_ethtool_opts(opts, iface)
            if tmp_ethtool:
                ethtool = {}
                for item in tmp_ethtool:
                    ethtool[_ETHTOOL_CONFIG_OPTS[item]] = tmp_ethtool[item]

                iface_data[addrfam]["ethtool"] = ethtool
                # return a list of sorted keys to ensure consistent order
                iface_data[addrfam]["ethtool_keys"] = sorted(ethtool)

        if iface_type == "bridge":
            bridging = _parse_bridge_opts(opts, iface)
            if bridging:
                iface_data[addrfam]["bridging"] = bridging
                iface_data[addrfam]["bridging_keys"] = sorted(bridging)
                iface_data[addrfam]["addrfam"] = addrfam

        elif iface_type == "bond":
            bonding = _parse_settings_bond(opts, iface)
            if bonding:
                iface_data[addrfam]["bonding"] = bonding
                iface_data[addrfam]["bonding"]["slaves"] = opts["slaves"]
                iface_data[addrfam]["bonding_keys"] = sorted(bonding)
                iface_data[addrfam]["addrfam"] = addrfam

        elif iface_type == "slave":
            adapters[iface]["master"] = opts["master"]

            opts["proto"] = "manual"
            iface_data[addrfam]["master"] = adapters[iface]["master"]
            iface_data[addrfam]["addrfam"] = addrfam

        elif iface_type == "vlan":
            iface_data[addrfam]["vlan_raw_device"] = re.sub(r"\.\d*", "", iface)
            iface_data[addrfam]["addrfam"] = addrfam

        elif iface_type == "pppoe":
            tmp_ethtool = _parse_ethtool_pppoe_opts(opts, iface)
            if tmp_ethtool:
                for item in tmp_ethtool:
                    adapters[iface]["data"][addrfam][
                        _DEB_CONFIG_PPPOE_OPTS[item]
                    ] = tmp_ethtool[item]
            iface_data[addrfam]["addrfam"] = addrfam

    opts.pop("mode", None)

    for opt, val in opts.items():
        inet = None
        if opt.startswith("ipv4"):
            opt = opt[4:]
            inet = "inet"
            iface_data["inet"]["addrfam"] = "inet"
        elif opt.startswith("ipv6"):
            iface_data["inet6"]["addrfam"] = "inet6"
            opt = opt[4:]
            inet = "inet6"
        elif opt in [
            "ipaddr",
            "address",
            "ipaddresses",
            "addresses",
            "gateway",
            "proto",
        ]:
            iface_data["inet"]["addrfam"] = "inet"
            inet = "inet"

        _opt = SALT_ATTR_TO_DEBIAN_ATTR_MAP.get(opt, opt)
        _debopt = _opt.replace("-", "_")

        for addrfam in ["inet", "inet6"]:
            (valid, value, errmsg) = _validate_interface_option(
                _opt, val, addrfam=addrfam
            )
            if not valid:
                continue
            if inet is None and _debopt not in iface_data[addrfam]:
                iface_data[addrfam][_debopt] = value
            elif inet == addrfam:
                iface_data[addrfam][_debopt] = value

    for opt in [
        "up_cmds",
        "pre_up_cmds",
        "post_up_cmds",
        "down_cmds",
        "pre_down_cmds",
        "post_down_cmds",
    ]:
        if opt in opts:
            iface_data["inet"][opt] = opts[opt]
            iface_data["inet6"][opt] = opts[opt]

    # Remove incomplete/disabled inet blocks
    for (addrfam, opt) in [("inet", "enable_ipv4"), ("inet6", "enable_ipv6")]:
        if opts.get(opt, None) is False:
            iface_data.pop(addrfam)
        elif iface_data[addrfam].get("addrfam", "") != addrfam:
            iface_data.pop(addrfam)

    return adapters


def _parse_settings_source(opts, iface_type, enabled, iface):
    """
    Filters given options and outputs valid settings for a
    network interface.
    """
    adapters = salt.utils.odict.OrderedDict()
    adapters[iface] = salt.utils.odict.OrderedDict()

    adapters[iface]["type"] = iface_type

    adapters[iface]["data"] = salt.utils.odict.OrderedDict()
    iface_data = adapters[iface]["data"]
    iface_data["sources"] = [opts["source"]]

    return adapters


def _parse_network_settings(opts, current):
    """
    Filters given options and outputs valid settings for
    the global network settings file.
    """
    # Normalize keys
    opts = {k.lower(): v for (k, v) in opts.items()}
    current = {k.lower(): v for (k, v) in current.items()}
    result = {}

    valid = _CONFIG_TRUE + _CONFIG_FALSE
    if "enabled" not in opts:
        try:
            opts["networking"] = current["networking"]
            _log_default_network("networking", current["networking"])
        except ValueError:
            _raise_error_network("networking", valid)
    else:
        opts["networking"] = opts["enabled"]

    if opts["networking"] in valid:
        if opts["networking"] in _CONFIG_TRUE:
            result["networking"] = "yes"
        elif opts["networking"] in _CONFIG_FALSE:
            result["networking"] = "no"
    else:
        _raise_error_network("networking", valid)

    if "hostname" not in opts:
        try:
            opts["hostname"] = current["hostname"]
            _log_default_network("hostname", current["hostname"])
        except ValueError:
            _raise_error_network("hostname", ["server1.example.com"])

    if opts["hostname"]:
        result["hostname"] = opts["hostname"]
    else:
        _raise_error_network("hostname", ["server1.example.com"])

    if "search" in opts:
        result["search"] = opts["search"]

    return result


def _parse_routes(iface, opts):
    """
    Filters given options and outputs valid settings for
    the route settings file.
    """
    # Normalize keys
    opts = {k.lower(): v for (k, v) in opts.items()}
    result = {}
    if "routes" not in opts:
        _raise_error_routes(iface, "routes", "List of routes")

    for opt in opts:
        result[opt] = opts[opt]

    return result


def _write_file(iface, data, folder, pattern):
    """
    Writes a file to disk
    """
    filename = os.path.join(folder, pattern.format(iface))
    if not os.path.exists(folder):
        msg = "{0} cannot be written. {1} does not exist"
        msg = msg.format(filename, folder)
        log.error(msg)
        raise AttributeError(msg)
    with salt.utils.files.flopen(filename, "w") as fout:
        fout.write(salt.utils.stringutils.to_str(data))
    return filename


def _write_file_routes(iface, data, folder, pattern):
    """
    Writes a file to disk
    """
    # ifup / ifdown is executing given folder via run-parts.
    # according to run-parts man-page, only filenames with this pattern are
    # executed: (^[a-zA-Z0-9_-]+$)

    # In order to make the routes file work for vlan interfaces
    # (default would have been in example /etc/network/if-up.d/route-bond0.12)
    # these dots in the iface name need to be replaced by underscores, so it
    # can be executed by run-parts
    iface = iface.replace(".", "_")

    filename = os.path.join(folder, pattern.format(iface))
    if not os.path.exists(folder):
        msg = "{0} cannot be written. {1} does not exist"
        msg = msg.format(filename, folder)
        log.error(msg)
        raise AttributeError(msg)
    with salt.utils.files.flopen(filename, "w") as fout:
        fout.write(salt.utils.stringutils.to_str(data))

    __salt__["file.set_mode"](filename, "0755")
    return filename


def _write_file_network(data, filename, create=False):
    """
    Writes a file to disk
    If file does not exist, only create if create
    argument is True
    """
    if not os.path.exists(filename) and not create:
        msg = "{0} cannot be written. {0} does not exist and create is setto False".format(
            filename
        )
        log.error(msg)
        raise AttributeError(msg)
    with salt.utils.files.flopen(filename, "w") as fout:
        fout.write(salt.utils.stringutils.to_str(data))


def _read_temp(data):
    """
    Return what would be written to disk
    """
    tout = io.StringIO()
    tout.write(data)
    tout.seek(0)
    output = tout.readlines()
    tout.close()

    return output


def _read_temp_ifaces(iface, data):
    """
    Return what would be written to disk for interfaces
    """
    try:
        template = JINJA.get_template("debian_eth.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template debian_eth.jinja")
        return ""

    ifcfg = template.render({"name": iface, "data": data})
    # Return as an array so the difflib works
    return [item + "\n" for item in ifcfg.split("\n")]


def _write_file_ifaces(iface, data, **settings):
    """
    Writes a file to disk
    """
    try:
        eth_template = JINJA.get_template("debian_eth.jinja")
        source_template = JINJA.get_template("debian_source.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template debian_eth.jinja")
        return ""

    # Read /etc/network/interfaces into a dict
    adapters = _parse_interfaces()
    # Apply supplied settings over on-disk settings
    adapters[iface] = data

    ifcfg = ""
    for adapter in adapters:
        if "type" in adapters[adapter] and adapters[adapter]["type"] == "source":
            tmp = source_template.render({"name": adapter, "data": adapters[adapter]})
        else:
            tmp = eth_template.render({"name": adapter, "data": adapters[adapter]})
        ifcfg = ifcfg + tmp
        if adapter == iface:
            saved_ifcfg = tmp

    _SEPARATE_FILE = False
    if "filename" in settings:
        if not settings["filename"].startswith("/"):
            filename = "{}/{}".format(_DEB_NETWORK_DIR, settings["filename"])
        else:
            filename = settings["filename"]
        _SEPARATE_FILE = True
    else:
        if "filename" in adapters[adapter]["data"]:
            filename = adapters[adapter]["data"]
        else:
            filename = _DEB_NETWORK_FILE

    if not os.path.exists(os.path.dirname(filename)):
        msg = "{0} cannot be written."
        msg = msg.format(os.path.dirname(filename))
        log.error(msg)
        raise AttributeError(msg)
    with salt.utils.files.flopen(filename, "w") as fout:
        if _SEPARATE_FILE:
            fout.write(salt.utils.stringutils.to_str(saved_ifcfg))
        else:
            fout.write(salt.utils.stringutils.to_str(ifcfg))

    # Return as an array so the difflib works
    return saved_ifcfg.split("\n")


def _write_file_ppp_ifaces(iface, data):
    """
    Writes a file to disk
    """
    try:
        template = JINJA.get_template("debian_ppp_eth.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template debian_ppp_eth.jinja")
        return ""

    adapters = _parse_interfaces()
    adapters[iface] = data

    ifcfg = ""
    tmp = template.render({"data": adapters[iface]})
    ifcfg = tmp + ifcfg

    filename = _DEB_PPP_DIR + "/" + adapters[iface]["data"]["inet"]["provider"]
    if not os.path.exists(os.path.dirname(filename)):
        msg = "{0} cannot be written."
        msg = msg.format(os.path.dirname(filename))
        log.error(msg)
        raise AttributeError(msg)
    with salt.utils.files.fopen(filename, "w") as fout:
        fout.write(salt.utils.stringutils.to_str(ifcfg))

    # Return as an array so the difflib works
    return filename


def build_bond(iface, **settings):
    """
    Create a bond script in /etc/modprobe.d with the passed settings
    and load the bonding kernel module.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_bond bond0 mode=balance-alb
    """
    deb_major = __grains__["osrelease"][:1]

    opts = _parse_settings_bond(settings, iface)
    try:
        template = JINJA.get_template("conf.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template conf.jinja")
        return ""
    data = template.render({"name": iface, "bonding": opts})

    if "test" in settings and settings["test"]:
        return _read_temp(data)

    _write_file(iface, data, _DEB_NETWORK_CONF_FILES, "{}.conf".format(iface))
    path = os.path.join(_DEB_NETWORK_CONF_FILES, "{}.conf".format(iface))
    if deb_major == "5":
        for line_type in ("alias", "options"):
            cmd = [
                "sed",
                "-i",
                "-e",
                r"/^{}\s{}.*/d".format(line_type, iface),
                "/etc/modprobe.conf",
            ]
            __salt__["cmd.run"](cmd, python_shell=False)
        __salt__["file.append"]("/etc/modprobe.conf", path)

    # Load kernel module
    __salt__["kmod.load"]("bonding")

    # install ifenslave
    __salt__["pkg.install"]("ifenslave")

    return _read_file(path)


def build_interface(iface, iface_type, enabled, **settings):
    """
    Build an interface script for a network interface.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_interface eth0 eth <settings>
    """

    iface_type = iface_type.lower()

    if iface_type not in _IFACE_TYPES:
        _raise_error_iface(iface, iface_type, _IFACE_TYPES)

    if iface_type == "slave":
        settings["slave"] = "yes"
        if "master" not in settings:
            msg = "master is a required setting for slave interfaces"
            log.error(msg)
            raise AttributeError(msg)

    elif iface_type == "vlan":
        settings["vlan"] = "yes"
        __salt__["pkg.install"]("vlan")

    elif iface_type == "pppoe":
        settings["pppoe"] = "yes"
        if not __salt__["pkg.version"]("ppp"):
            inst = __salt__["pkg.install"]("ppp")

    elif iface_type == "bond":
        if "slaves" not in settings:
            msg = "slaves is a required setting for bond interfaces"
            log.error(msg)
            raise AttributeError(msg)

    elif iface_type == "bridge":
        if "ports" not in settings:
            msg = (
                "ports is a required setting for bridge interfaces on Debian "
                "or Ubuntu based systems"
            )
            log.error(msg)
            raise AttributeError(msg)
        __salt__["pkg.install"]("bridge-utils")

    if iface_type in ["eth", "bond", "bridge", "slave", "vlan", "pppoe"]:
        opts = _parse_settings_eth(settings, iface_type, enabled, iface)

    if iface_type in ["source"]:
        opts = _parse_settings_source(settings, iface_type, enabled, iface)

    if "test" in settings and settings["test"]:
        return _read_temp_ifaces(iface, opts[iface])

    ifcfg = _write_file_ifaces(iface, opts[iface], **settings)

    if iface_type == "pppoe":
        _write_file_ppp_ifaces(iface, opts[iface])

    # ensure lines in list end with newline, so difflib works
    return [item + "\n" for item in ifcfg]


def build_routes(iface, **settings):
    """
    Add route scripts for a network interface using up commands.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_routes eth0 <settings>
    """

    opts = _parse_routes(iface, settings)
    try:
        template = JINJA.get_template("route_eth.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template route_eth.jinja")
        return ""

    add_routecfg = template.render(route_type="add", routes=opts["routes"], iface=iface)

    del_routecfg = template.render(route_type="del", routes=opts["routes"], iface=iface)

    if "test" in settings and settings["test"]:
        return _read_temp(add_routecfg + del_routecfg)

    filename = _write_file_routes(iface, add_routecfg, _DEB_NETWORK_UP_DIR, "route-{0}")
    results = _read_file(filename)

    filename = _write_file_routes(
        iface, del_routecfg, _DEB_NETWORK_DOWN_DIR, "route-{0}"
    )
    results += _read_file(filename)

    return results


def down(iface, iface_type):
    """
    Shutdown a network interface

    CLI Example:

    .. code-block:: bash

        salt '*' ip.down eth0 eth
    """
    # Slave devices are controlled by the master.
    # Source 'interfaces' aren't brought down.
    if iface_type not in ["slave", "source"]:
        return __salt__["cmd.run"](["ifdown", iface])
    return None


def get_bond(iface):
    """
    Return the content of a bond script

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_bond bond0
    """
    path = os.path.join(_DEB_NETWORK_CONF_FILES, "{}.conf".format(iface))
    return _read_file(path)


def get_interface(iface):
    """
    Return the contents of an interface script

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_interface eth0
    """

    adapters = _parse_interfaces()
    if iface in adapters:
        try:
            if iface == "source":
                template = JINJA.get_template("debian_source.jinja")
            else:
                template = JINJA.get_template("debian_eth.jinja")
        except jinja2.exceptions.TemplateNotFound:
            log.error("Could not load template debian_eth.jinja")
            return ""

        ifcfg = template.render({"name": iface, "data": adapters[iface]})

        # ensure lines in list end with newline, so difflib works
        return [item + "\n" for item in ifcfg.split("\n")]
    else:
        return []


def up(iface, iface_type):  # pylint: disable=C0103
    """
    Start up a network interface

    CLI Example:

    .. code-block:: bash

        salt '*' ip.up eth0 eth
    """
    # Slave devices are controlled by the master.
    # Source 'interfaces' aren't brought up.
    if iface_type not in ("slave", "source"):
        return __salt__["cmd.run"](["ifup", iface])
    return None


def get_network_settings():
    """
    Return the contents of the global network script.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_network_settings
    """
    skip_etc_default_networking = (
        __grains__["osfullname"] == "Ubuntu"
        and int(__grains__["osrelease"].split(".")[0]) >= 12
    )

    if skip_etc_default_networking:
        settings = {}
        if __salt__["service.available"]("networking"):
            if __salt__["service.status"]("networking"):
                settings["networking"] = "yes"
            else:
                settings["networking"] = "no"
        else:
            settings["networking"] = "no"

        hostname = _parse_hostname()
        domainname = _parse_domainname()

        settings["hostname"] = hostname
        settings["domainname"] = domainname

    else:
        settings = _parse_current_network_settings()

    try:
        template = JINJA.get_template("display-network.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template display-network.jinja")
        return ""

    network = template.render(settings)
    return _read_temp(network)


def get_routes(iface):
    """
    Return the routes for the interface

    CLI Example:

    .. code-block:: bash

        salt '*' ip.get_routes eth0
    """

    filename = os.path.join(_DEB_NETWORK_UP_DIR, "route-{}".format(iface))
    results = _read_file(filename)

    filename = os.path.join(_DEB_NETWORK_DOWN_DIR, "route-{}".format(iface))
    results += _read_file(filename)

    return results


def apply_network_settings(**settings):
    """
    Apply global network configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.apply_network_settings
    """
    if "require_reboot" not in settings:
        settings["require_reboot"] = False

    if "apply_hostname" not in settings:
        settings["apply_hostname"] = False

    hostname_res = True
    if settings["apply_hostname"] in _CONFIG_TRUE:
        if "hostname" in settings:
            hostname_res = __salt__["network.mod_hostname"](settings["hostname"])
        else:
            log.warning(
                "The network state sls is trying to apply hostname "
                "changes but no hostname is defined."
            )
            hostname_res = False

    res = True
    if settings["require_reboot"] in _CONFIG_TRUE:
        log.warning(
            "The network state sls is requiring a reboot of the system to "
            "properly apply network configuration."
        )
        res = True
    else:
        stop = __salt__["service.stop"]("networking")
        time.sleep(2)
        res = stop and __salt__["service.start"]("networking")

    return hostname_res and res


def build_network_settings(**settings):
    """
    Build the global network script.

    CLI Example:

    .. code-block:: bash

        salt '*' ip.build_network_settings <settings>
    """
    changes = []

    # Read current configuration and store default values
    current_network_settings = _parse_current_network_settings()

    # Build settings
    opts = _parse_network_settings(settings, current_network_settings)

    # Ubuntu has moved away from /etc/default/networking
    # beginning with the 12.04 release so we disable or enable
    # the networking related services on boot
    skip_etc_default_networking = (
        __grains__["osfullname"] == "Ubuntu"
        and int(__grains__["osrelease"].split(".")[0]) >= 12
    )
    if skip_etc_default_networking:
        if opts["networking"] == "yes":
            service_cmd = "service.enable"
        else:
            service_cmd = "service.disable"

        if __salt__["service.available"]("NetworkManager"):
            __salt__[service_cmd]("NetworkManager")

        if __salt__["service.available"]("networking"):
            __salt__[service_cmd]("networking")
    else:
        try:
            template = JINJA.get_template("network.jinja")
        except jinja2.exceptions.TemplateNotFound:
            log.error("Could not load template network.jinja")
            return ""
        network = template.render(opts)

        if "test" in settings and settings["test"]:
            return _read_temp(network)
        # Write settings
        _write_file_network(network, _DEB_NETWORKING_FILE, True)

    # Get hostname and domain from opts
    sline = opts["hostname"].split(".", 1)
    opts["hostname"] = sline[0]
    current_domainname = current_network_settings["domainname"]
    current_searchdomain = current_network_settings["searchdomain"]

    new_domain = False
    if len(sline) > 1:
        new_domainname = sline[1]
        if new_domainname != current_domainname:
            domainname = new_domainname
            opts["domainname"] = new_domainname
            new_domain = True
        else:
            domainname = current_domainname
            opts["domainname"] = domainname
    else:
        domainname = current_domainname
        opts["domainname"] = domainname

    new_search = False
    if "search" in opts:
        new_searchdomain = opts["search"]
        if new_searchdomain != current_searchdomain:
            searchdomain = new_searchdomain
            opts["searchdomain"] = new_searchdomain
            new_search = True
        else:
            searchdomain = current_searchdomain
            opts["searchdomain"] = searchdomain
    else:
        searchdomain = current_searchdomain
        opts["searchdomain"] = searchdomain

    # If the domain changes, then we should write the resolv.conf file.
    if new_domain or new_search:
        # Look for existing domain line and update if necessary
        resolve = _parse_resolve()
        domain_prog = re.compile(r"domain\s+")
        search_prog = re.compile(r"search\s+")
        new_contents = []

        for item in _read_file(_DEB_RESOLV_FILE):
            if domain_prog.match(item):
                item = "domain {}".format(domainname)
            elif search_prog.match(item):
                item = "search {}".format(searchdomain)
            new_contents.append(item)

        # A domain line didn't exist so we'll add one in
        # with the new domainname
        if "domain" not in resolve:
            new_contents.insert(0, "domain {}".format(domainname))

        # A search line didn't exist so we'll add one in
        # with the new search domain
        if "search" not in resolve:
            new_contents.insert("domain" in resolve, "search {}".format(searchdomain))

        new_resolv = "\n".join(new_contents)

        # Write /etc/resolv.conf
        if not ("test" in settings and settings["test"]):
            _write_file_network(new_resolv, _DEB_RESOLV_FILE)

    #  used for returning the results back
    try:
        template = JINJA.get_template("display-network.jinja")
    except jinja2.exceptions.TemplateNotFound:
        log.error("Could not load template display-network.jinja")
        return ""
    network = template.render(opts)
    changes.extend(_read_temp(network))

    return changes



"""
This is the default pcre matcher.
"""

import re


def match(tgt, opts=None, minion_id=None):
    """
    Returns true if the passed pcre regex matches
    """
    if not opts:
        opts = __opts__
    if not minion_id:
        minion_id = opts.get("id")

    return bool(re.match(tgt, minion_id))



import json

import pytest

from salt.utils.versions import Version
from tests.support.helpers import SaltVirtualEnv
from tests.support.pytest.helpers import FakeSaltExtension

pytestmark = [
    # These are slow because they create a virtualenv and install salt in it
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def salt_extension(tmp_path_factory):
    with FakeSaltExtension(
        tmp_path_factory=tmp_path_factory, name="salt-ext-loader-test"
    ) as extension:
        yield extension


@pytest.fixture
def venv(tmp_path):
    with SaltVirtualEnv(venv_dir=tmp_path / ".venv") as _venv:
        yield _venv


def test_new_entry_points_passing_module(venv, salt_extension, salt_minion_factory):
    # Install our extension into the virtualenv
    venv.install(str(salt_extension.srcdir))
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages
    code = """
    import sys
    import json
    import salt._logging
    import salt.loader

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    loader = salt.loader.minion_mods(minion_config)
    print(json.dumps(list(loader)))
    """
    ret = venv.run_code(code, input=json.dumps(salt_minion_factory.config.copy()))
    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echo1" in loader_functions
    assert "foobar.echo2" in loader_functions


def test_new_entry_points_passing_func_returning_a_dict(
    venv, salt_extension, salt_minion_factory
):
    # Install our extension into the virtualenv
    venv.install(str(salt_extension.srcdir))
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages
    code = """
    import sys
    import json
    import salt._logging
    import salt.loader

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    loader = salt.loader.runner(minion_config)
    print(json.dumps(list(loader)))
    """
    ret = venv.run_code(code, input=json.dumps(salt_minion_factory.config.copy()))
    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echo1" in loader_functions
    assert "foobar.echo2" in loader_functions


def test_old_entry_points_yielding_paths(venv, salt_extension, salt_minion_factory):
    # Install our extension into the virtualenv
    venv.install(str(salt_extension.srcdir))
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages
    code = """
    import sys
    import json
    import salt._logging
    import salt.loader

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    functions = salt.loader.minion_mods(minion_config)
    utils = salt.loader.utils(minion_config)
    serializers = salt.loader.serializers(minion_config)
    loader = salt.loader.states(minion_config, functions, utils, serializers)
    print(json.dumps(list(loader)))
    """
    ret = venv.run_code(code, input=json.dumps(salt_minion_factory.config.copy()))
    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echoed" in loader_functions


def test_utils_loader_does_not_load_extensions(
    venv, salt_extension, salt_minion_factory
):
    # Install our extension into the virtualenv
    venv.install(str(salt_extension.srcdir))
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages
    code = """
    import sys
    import json
    import salt._logging
    import salt.loader

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    loader = salt.loader.utils(minion_config)
    print(json.dumps(list(loader)))
    """
    ret = venv.run_code(code, input=json.dumps(salt_minion_factory.config.copy()))
    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echo" not in loader_functions


def test_extension_discovery_without_reload_with_importlib_metadata_installed(
    venv, salt_extension, salt_minion_factory
):
    # Install our extension into the virtualenv
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name not in installed_packages
    venv.install("importlib-metadata==4.6.4")
    code = """
    import sys
    import json
    import subprocess
    import salt._logging
    import salt.loader

    extension_path = "{}"

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    loader = salt.loader.minion_mods(minion_config)

    if "foobar.echo1" in loader:
        sys.exit(1)

    # Install the extension
    proc = subprocess.run(
        [sys.executable, "-m", "pip", "install", extension_path],
        check=False,
        shell=False,
        stdout=subprocess.PIPE,
    )
    if proc.returncode != 0:
        sys.exit(2)

    loader = salt.loader.minion_mods(minion_config)
    if "foobar.echo1" not in loader:
        sys.exit(3)

    print(json.dumps(list(loader)))
    """.format(
        salt_extension.srcdir
    )
    ret = venv.run_code(
        code, input=json.dumps(salt_minion_factory.config.copy()), check=False
    )
    # Exitcode 1 - Extension was already installed
    # Exitcode 2 - Failed to install the extension
    # Exitcode 3 - Extension was not found within the same python process after being installed
    assert ret.returncode == 0
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages

    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echo1" in loader_functions
    assert "foobar.echo2" in loader_functions


def test_extension_discovery_without_reload_with_importlib_metadata_5_installed(
    venv, salt_extension, salt_minion_factory
):
    # Install our extension into the virtualenv
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name not in installed_packages
    venv.install("importlib-metadata>=3.3.0")
    code = """
    import sys
    import json
    import subprocess
    import salt._logging
    import salt.loader

    extension_path = "{}"

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    loader = salt.loader.minion_mods(minion_config)

    if "foobar.echo1" in loader:
        sys.exit(1)

    # Install the extension
    proc = subprocess.run(
        [sys.executable, "-m", "pip", "install", extension_path],
        check=False,
        shell=False,
        stdout=subprocess.PIPE,
    )
    if proc.returncode != 0:
        sys.exit(2)

    loader = salt.loader.minion_mods(minion_config)
    if "foobar.echo1" not in loader:
        sys.exit(3)

    print(json.dumps(list(loader)))
    """.format(
        salt_extension.srcdir
    )
    ret = venv.run_code(
        code, input=json.dumps(salt_minion_factory.config.copy()), check=False
    )
    # Exitcode 1 - Extension was already installed
    # Exitcode 2 - Failed to install the extension
    # Exitcode 3 - Extension was not found within the same python process after being installed
    assert ret.returncode == 0
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages

    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echo1" in loader_functions
    assert "foobar.echo2" in loader_functions


def test_extension_discovery_without_reload_with_bundled_importlib_metadata(
    venv, salt_extension, salt_minion_factory
):
    # Install our extension into the virtualenv
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name not in installed_packages
    if "importlib-metadata" in installed_packages:
        importlib_metadata_version = installed_packages["importlib-metadata"]
        if Version(importlib_metadata_version) >= Version("3.3.0"):
            venv.install("-U", "importlib-metadata<3.3.0")
    code = """
    import sys
    import json
    import subprocess
    import salt._logging
    import salt.loader

    extension_path = "{}"

    minion_config = json.loads(sys.stdin.read())
    salt._logging.set_logging_options_dict(minion_config)
    salt._logging.setup_logging()
    loader = salt.loader.minion_mods(minion_config)
    if "foobar.echo1" in loader:
        sys.exit(1)

    # Install the extension
    proc = subprocess.run(
        [sys.executable, "-m", "pip", "install", extension_path],
        check=False,
        shell=False,
        stdout=subprocess.PIPE,
    )
    if proc.returncode != 0:
        sys.exit(2)

    loader = salt.loader.minion_mods(minion_config)
    if "foobar.echo1" not in loader:
        sys.exit(3)

    print(json.dumps(list(loader)))
    """.format(
        salt_extension.srcdir
    )
    ret = venv.run_code(
        code, input=json.dumps(salt_minion_factory.config.copy()), check=False
    )
    # Exitcode 1 - Extension was already installed
    # Exitcode 2 - Failed to install the extension
    # Exitcode 3 - Extension was not found within the same python process after being installed
    assert ret.returncode == 0
    installed_packages = venv.get_installed_packages()
    assert salt_extension.name in installed_packages

    loader_functions = json.loads(ret.stdout)

    # A non existing module should not appear in the loader
    assert "monty.python" not in loader_functions

    # But our extension's modules should appear on the loader
    assert "foobar.echo1" in loader_functions
    assert "foobar.echo2" in loader_functions
